#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import datetime
from x509sak.ConsolePrinter import ConsolePrinter
from x509sak.estimate.Judgement import SecurityJudgements
from x509sak.estimate import Verdict, Commonness, Compatibility, StandardDeviationType

class AnalysisPrinter():
	def __init__(self, outfile, analyses, **kwargs):
		self._outfile = outfile
		self._analyses = analyses
		self._kwargs = kwargs

	def print(self, analyses):
		raise NotImplementedError("Method not overriden.")

class AnalysisPrinterText(AnalysisPrinter):
	def _create_printer(self, use_ansi):
		if use_ansi:
			return ConsolePrinter(self._outfile).add_subs({
				"<good>":		"\x1b[32m",
				"<warn>":		"\x1b[33m",
				"<error>":		"\x1b[31m",
				"<insecure>":	"\x1b[41m",
				"<end>":		"\x1b[0m",
			})
		else:
			return ConsolePrinter(self._outfile).add_subs({
				"<good>":		"",
				"<warn>":		"",
				"<error>":		"",
				"<insecure>":	"",
				"<end>":		"",
			})

	@staticmethod
	def _fmt_textual_verdict(judgement):
		textual_verdict = [ ]
		if judgement.bits is not None:
			textual_verdict.append("%d bits" % (judgement.bits))
		if judgement.verdict is not None:
			textual_verdict.append({
				Verdict.BEST_IN_CLASS:					"best-in-class security",
				Verdict.HIGH:							"high security",
				Verdict.MEDIUM:							"medium security",
				Verdict.WEAK:							"weak security",
				Verdict.BROKEN:							"broken security",
				Verdict.NO_SECURITY:					"no security",
			}[judgement.verdict])
		if judgement.commonness is not None:
			textual_verdict.append({
				Commonness.COMMON:						"common",
				Commonness.FAIRLY_COMMON:				"fairly common",
				Commonness.UNUSUAL:						"unusual",
				Commonness.HIGHLY_UNUSUAL:				"highly unusual",
			}[judgement.commonness])
		if judgement.compatibility is not None:
			textual_verdict.append({
				Compatibility.FULLY_COMPLIANT:			"fully standards compliant",
				Compatibility.LIMITED_SUPPORT:			"limited support",
				Compatibility.STANDARDS_DEVIATION:		"standards deviation",
			}[judgement.compatibility])

		if len(textual_verdict) == 0:
			return None
		else:
			return ", ".join(textual_verdict)

	@staticmethod
	def _fmt_color(judgement):
		color = "end"
		if judgement.verdict in [ Verdict.BEST_IN_CLASS, Verdict.HIGH ]:
			color = "good"
		if (judgement.verdict == Verdict.MEDIUM) or (judgement.commonness == Commonness.UNUSUAL) or (judgement.compatibility == Compatibility.LIMITED_SUPPORT):
			color = "warn"
		if (judgement.compatibility == Compatibility.STANDARDS_DEVIATION) and (judgement.standard is not None):
			if judgement.standard.deviation_type == StandardDeviationType.RECOMMENDATION:
				color = "warn"
			elif judgement.standard.deviation_type == StandardDeviationType.VIOLATION:
				color = "error"
			elif judgement.standard.deviation_type is None:
				# Just informative
				color = "end"
			else:
				raise NotImplementedError(judgement.standard.deviation_type)
		if (judgement.verdict in [ Verdict.WEAK, Verdict.BROKEN ]) or (judgement.commonness == Commonness.HIGHLY_UNUSUAL):
			color = "error"
		if (judgement.verdict == Verdict.NO_SECURITY) or (judgement.compatibility == Compatibility.STANDARDS_DEVIATION):
			color = "insecure"
		return color

	def _fmt_security_judgement(self, judgement):
		color = self._fmt_color(judgement)
		text = "%s: %s" % (judgement.code.topic, judgement.text)

		if (judgement.compatibility == Compatibility.STANDARDS_DEVIATION) and (judgement.standard is not None):
			if judgement.standard.deviation_type == StandardDeviationType.RECOMMENDATION:
				text += " This goes against the recommendation of %s." % (judgement.standard)
			elif judgement.standard.deviation_type == StandardDeviationType.VIOLATION:
				text += " This is in violation of %s." % (judgement.standard)
			elif judgement.standard.deviation_type is None:
				text += " This is an informative reference."
			else:
				raise NotImplementedError(judgement.standard.deviation_type)

		textual_verdict = self._fmt_textual_verdict(judgement)
		if textual_verdict is not None:
			text = "%s {%s}" % (text, textual_verdict)
		return "<%s>%s<end>" % (color, text)

	def _print_security_judgements(self, judgements_data, indent = ""):
		judgements = SecurityJudgements.from_dict(judgements_data)
		if len(judgements) == 0:
			self._printer.print("%sNo comments regarding this check." % (indent))
		elif len(judgements) == 1:
			self._printer.print("%s%s" % (indent, self._fmt_security_judgement(judgements[0])))
		else:
			for (jid, judgement) in enumerate(judgements, 1):
				self._printer.print("%s%d / %d: %s" % (indent, jid, len(judgements), self._fmt_security_judgement(judgement)))

			summary_judgement = judgements.summary_judgement()
			color = self._fmt_color(summary_judgement)
			self._printer.print("%s    -> Summary: <%s>%s<end>" % (indent, color, self._fmt_textual_verdict(summary_judgement)))

	@staticmethod
	def _fmt_time(isotime):
		ts = datetime.datetime.strptime(isotime, "%Y-%m-%dT%H:%M:%SZ")
		return ts.strftime("%c UTC")

	def _fmt_validity(self, validity):
		if validity is None:
			return "error parsing timestamp"
		else:
			return self._fmt_time(validity["iso"])

	def _print_analysis(self, analysis):
		self._printer.heading("Metadata")
		if analysis["source"]["certs_total"] == 1:
			self._printer.print("Source : %s" % (analysis["source"]["name"]))
		else:
			self._printer.print("Source : %s (certificate %d of %d)" % (analysis["source"]["name"], analysis["source"]["cert_no"], analysis["source"]["certs_total"]))
		self._printer.print("Issuer : %s" % (analysis["issuer"]["pretty"]))
		if len(analysis["issuer"]["security"]["components"]) > 0:
			self._print_security_judgements(analysis["issuer"]["security"], indent = "  ")
		self._printer.print("Subject: %s" % (analysis["subject"]["pretty"]))
		if len(analysis["subject"]["security"]["components"]) > 0:
			self._print_security_judgements(analysis["subject"]["security"], indent = "  ")
		if len(analysis["security"]["components"]) > 0:
			self._printer.print("Misc observations:")
			self._print_security_judgements(analysis["security"], indent = "  ")
		self._printer.print()

		self._printer.heading("Validity")
		self._printer.print("Valid from : %s" % (self._fmt_validity(analysis["validity"]["not_before"])))
		self._printer.print("Valid until: %s" % (self._fmt_validity(analysis["validity"]["not_after"])))
		self._printer.print("Lifetime   : %.1f years" % (analysis["validity"]["validity_days"] / 365))
		self._print_security_judgements(analysis["validity"]["security"], indent = "  ")
		self._printer.print()

		if "ca" in analysis:
			self._printer.heading("CA Certificate")
			self._printer.print("Subject: %s" % (analysis["ca"]["ca_subject"]["pretty"]))
			self._print_security_judgements(analysis["ca"]["security"], indent = "  ")
			self._printer.print()

		self._printer.heading("Public Key")
		self._printer.print("Used cryptography: %s" % (analysis["pubkey"]["pretty"]))
		self._print_security_judgements(analysis["pubkey"]["security"], indent = "    ")
		self._printer.print()

		self._printer.heading("Signature")
		self._printer.print("Signature algorithm: %s" % (analysis["signature"]["pretty"]))
		self._print_security_judgements(analysis["signature"]["security"], indent = "    ")
		if "hash_fnc" in analysis["signature"]:
			self._printer.print("Hash function      : %s" % (analysis["signature"]["hash_fnc"]["name"]))
			self._print_security_judgements(analysis["signature"]["hash_fnc"]["security"], indent = "    ")
		if "sig_fnc" in analysis["signature"]:
			self._printer.print("Signature function : %s" % (analysis["signature"]["sig_fnc"]["name"]))
			self._print_security_judgements(analysis["signature"]["sig_fnc"]["security"], indent = "    ")
		self._printer.print()

		self._printer.heading("X.509 Extensions")
		self._print_security_judgements(analysis["extensions"]["security"], indent = "    ")
		self._printer.print()

		if len(analysis["purpose"]) > 0:
			self._printer.heading("Certificate Purpose")
			for purpose_check in analysis["purpose"]:
				if purpose_check["type"] == "name_match":
					self._printer.print("   Name match '%s':" % (purpose_check["name"]))
					self._print_security_judgements(purpose_check["security"], indent = "       ")
				elif purpose_check["type"] == "purpose_match":
					self._printer.print("   Purpose %s:" % (purpose_check["purpose"]))
					self._print_security_judgements(purpose_check["security"], indent = "       ")
				else:
					raise NotImplementedError(purpose_check["type"])
			self._printer.print()

	def print(self, use_ansi = False):
		self._printer = self._create_printer(use_ansi)
		for analysis in self._analyses["data"]:
			self._print_analysis(analysis)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import math
import hashlib
import pyasn1.codec.der.decoder
from x509sak.KwargsChecker import KwargsChecker
from x509sak.NumberTheory import NumberTheory
from x509sak.Exceptions import InvalidInputException, UnsupportedEncodingException
from x509sak.ASN1Models import ECFieldParametersPrimeField, ECFieldParametersCharacteristicTwoField, ECFieldParametersCharacteristicTwoFieldTrinomial, ECFieldParametersCharacteristicTwoFieldPentanomial
from x509sak.Tools import ASN1Tools
from x509sak.OID import OID, OIDDB

class EllipticCurvePoint():
	"""Affine representation of an ECC curve point."""
	def __init__(self, curve, x, y):
		self._curve = curve
		self._x = x
		self._y = y

	def on_curve(self):
		return self._curve.on_curve(self)

	def encode(self):
		return self._curve.encode_point(self)

	@property
	def curve(self):
		return self._curve

	@property
	def x(self):
		return self._x

	@property
	def y(self):
		return self._y

	def scalar_mul(self, scalar):
		result = self.curve.neutral_point
		accumulator = self
		for bit in range(scalar.bit_length()):
			if ((scalar >> bit) & 1) == 1:
				# Bit set, add
				result = self.curve.point_addition(result, accumulator)
			accumulator = self.curve.point_addition(accumulator, accumulator)
		return result

	def __eq__(self, other):
		return (other is not None) and ((self.curve, self.x, self.y) == (other.curve, other.x, other.y))

	def __str__(self):
		return "(0x%x, 0x%x) on %s" % (self.x, self.y, self.curve)

class EllipticCurve():
	_CURVE_TYPE = None
	_DomainArgs = None
	_CurveTypes = { }

	def __init__(self, metadata = None, **domain_parameters):
		if self._DomainArgs is not None:
			self._DomainArgs.check(domain_parameters, "EllipticCurve")
		self._domain_parameters = domain_parameters
		self._metadata = metadata
		if self._metadata is None:
			self._metadata = { }

	@classmethod
	def register(cls, ec_class):
		cls._CurveTypes[ec_class._CURVE_TYPE] = ec_class
		return ec_class

	@classmethod
	def get_class_for_curvetype(cls, curve_type):
		return cls._CurveTypes.get(curve_type)

	@property
	def curvetype(self):
		if self._CURVE_TYPE is None:
			raise NotImplementedError(self.__class__.__name__)
		return self._CURVE_TYPE

	@property
	def name(self):
		return self._metadata.get("name")

	@property
	def oid(self):
		return self._metadata.get("oid")

	@property
	def domain_parameters(self):
		return dict(self._domain_parameters)

	@property
	def order_bits(self):
		raise NotImplementedError(self.__class__.__name__)

	@property
	def field_bits(self):
		raise NotImplementedError(self.__class__.__name__)

	@property
	def element_octet_cnt(self):
		return (self.field_bits + 7) // 8

	def point(self, x, y):
		return EllipticCurvePoint(self, x, y)

	@property
	def G(self):
		return self.point(self.Gx, self.Gy)

	@property
	def neutral_point(self):
		raise NotImplementedError(self.__class__.__name__)

	def on_curve(self, point):
		raise NotImplementedError(self.__class__.__name__)

	def point_addition(self, P, Q):
		raise NotImplementedError(self.__class__.__name__)

	def encode_point(self, point):
		return bytes([ 0x04 ]) + point.x.to_bytes(length = self.element_octet_cnt, byteorder = "big") + point.y.to_bytes(length = self.element_octet_cnt, byteorder = "big")

	def decode_point(self, serialized_point):
		if serialized_point[0] == 0x04:
			expected_length = 1 + (2 * self.element_octet_cnt)
			if len(serialized_point) == expected_length:
				Gx = int.from_bytes(serialized_point[1 : 1 + self.element_octet_cnt], byteorder = "big")
				Gy = int.from_bytes(serialized_point[1 + self.element_octet_cnt : 1 + (2 * self.element_octet_cnt)], byteorder = "big")
				return EllipticCurvePoint(x = Gx, y = Gy, curve = self)
			else:
				raise InvalidInputException("Do not know how to decode explicit serialized point with length %d (expected %d = 1 + 2 * %d bytes)." % (len(serialized_point), 1 + (2 * self.element_octet_cnt), self.element_octet_cnt))
		else:
			raise UnsupportedEncodingException("Do not know how to decode serialized point in non-explicit point format 0x%x." % (serialized_point[0]))

	@classmethod
	def from_asn1(cls, asn1):
		"""Decode explicitly encoded elliptic curve domain parameters, given as a Sequence (SpecifiedECDomain)."""
		version = int(asn1["version"])
		if version != 1:
			raise InvalidInputException("Attempted to decode the excplicit EC domain and saw unknown version %d." % (version))

		field_type = OID.from_asn1(asn1["fieldID"]["fieldType"])
		field_type_id = OIDDB.ECFieldType.get(field_type)
		if field_type_id is None:
			raise InvalidInputException("Encountered explicit EC domain parameters in unknown field with OID %s." % (str(field_type)))

		domain_parameters = {
			"a":	int.from_bytes(bytes(asn1["curve"]["a"]), byteorder = "big"),
			"b":	int.from_bytes(bytes(asn1["curve"]["b"]), byteorder = "big"),
			"n":	int(asn1["order"]),
		}
		if asn1["cofactor"].hasValue():
			domain_parameters["h"] = int(asn1["cofactor"])
		base_point = bytes(asn1["base"])
		if field_type_id == "prime-field":
			(field_params, tail) = pyasn1.codec.der.decoder.decode(bytes(asn1["fieldID"]["parameters"]), asn1Spec = ECFieldParametersPrimeField())
			if len(tail) != 0:
				raise InvalidInputException("Attempted to decode the excplicit EC domain and encountered %d bytes of trailing data of the prime basis Integer." % (len(tail)))

			domain_parameters.update({
				"p":	int(field_params),
			})
			return cls.get_class_for_curvetype("prime").instantiate(domain_parameters, base_point)
		elif field_type_id == "characteristic-two-field":
			(field_params, tail) = pyasn1.codec.der.decoder.decode(bytes(asn1["fieldID"]["parameters"]), asn1Spec = ECFieldParametersCharacteristicTwoField())
			if len(tail) != 0:
				raise InvalidInputException("Attempted to decode the excplicit EC domain and encountered %d bytes of trailing data of the characteristic two field Sequence." % (len(tail)))

			basis_type = OID.from_asn1(field_params["basis"])
			basis_type_id = OIDDB.ECTwoFieldBasistype.get(basis_type)
			if basis_type_id is None:
				raise InvalidInputException("Unknown two-field basis type with OID %s found in public key." % (str(basis_type)))

			# Field width is common to all two-fields
			domain_parameters.update({
				"m":		int(field_params["m"]),
				"basis":	basis_type_id,
			})

			if basis_type_id == "gnBasis":
				raise InvalidInputException("Binary field explicit domain parameters with Gaussian polynomial basis is not implemented.")
			elif basis_type_id == "tpBasis":
				(params, tail) = ASN1Tools.redecode(field_params["parameters"], ECFieldParametersCharacteristicTwoFieldTrinomial())
				if len(tail) != 0:
					raise InvalidInputException("Attempted to decode the excplicit EC domain and encountered %d bytes of trailing data of the characteristic two field trinomial basis." % (len(tail)))
				poly = [ domain_parameters["m"], int(params), 0 ]
			elif basis_type_id == "ppBasis":
				(params, tail) = ASN1Tools.redecode(field_params["parameters"], ECFieldParametersCharacteristicTwoFieldPentanomial())
				if len(tail) != 0:
					raise InvalidInputException("Attempted to decode the excplicit EC domain and encountered %d bytes of trailing data of the characteristic two field pentanomial basis." % (len(tail)))
				poly = [ domain_parameters["m"], int(params["k1"]), int(params["k2"]), int(params["k3"]), 0 ]
			else:
				raise NotImplementedError("Binary field basis", basis_type_id)
			domain_parameters.update({
				"m":		int(field_params["m"]),
				"poly":		poly,
			})
			return cls.get_class_for_curvetype("binary").instantiate(domain_parameters, base_point)
		else:
			raise NotImplementedError("Explicit EC domain parameter encoding for field type \"%s\" is not implemented." % (field_type_id))

	@classmethod
	def instantiate(cls, domain_parameters, encoded_base_point):
		curve_without_generator = cls(**domain_parameters)
		G = curve_without_generator.decode_point(encoded_base_point)
		domain_parameters.update({
			"Gx":	G.x,
			"Gy":	G.y,
		})
		curve = cls(**domain_parameters)
		return curve

	def __hash__(self):
		return hash(self.characteristic_cmpkey)

	def __eq__(self, other):
		return self.characteristic_cmpkey == other.characteristic_cmpkey

	def __lt__(self, other):
		return self.characteristic_cmpkey < other.characteristic_cmpkey

	def __neq__(self, other):
		return not (self == other)

	@property
	def characteristic_cmpkey(self):
		raise NotImplementedError(self.__class__.__name__)

	def __getattr__(self, key):
		if key in self._domain_parameters:
			return self._domain_parameters[key]
		else:
			if self._DomainArgs.is_optional_argument(key):
				return None
			else:
				raise AttributeError(key)

	def __str__(self):
		if self.name is None:
			return "Order %d %s" % (self.n.bit_length(), self.__class__.__name__)
		else:
			return "%s %s" % (self.__class__.__name__, self.name)

@EllipticCurve.register
class PrimeFieldEllipticCurve(EllipticCurve):
	"""y^2 = x^3 + ax + b (mod p)"""
	_DomainArgs = KwargsChecker(required_arguments = set([ "p", "a", "b", "n" ]), optional_arguments = set([ "h", "Gx", "Gy" ]))
	_CURVE_TYPE = "prime"

	@property
	def neutral_point(self):
		return None

	def point_addition(self, P, Q):
		if P is None:
			# O + Q = Q
			return Q
		elif Q is None:
			# P + O = P
			return P
		elif ((P.x == Q.x) and (((P.y + Q.y) % self.p) == 0)):
			# P + (-P) = O
			return None
		elif P == Q:
			# Point doubling
			s = ((3 * P.x ** 2) + self.a) * NumberTheory.modinv(2 * P.y, self.p)
			x = (s * s - (2 * P.x)) % self.p
			y = (s * (P.x - x) - P.y) % self.p
			return EllipticCurvePoint(self, x, y)
		else:
			# Point addition
			s = (P.y - Q.y) * NumberTheory.modinv(P.x - Q.x, self.p)
			x = ((s ** 2) - P.x - Q.x) % self.p
			y = (s * (P.x - x) - P.y) % self.p
			return EllipticCurvePoint(self, x, y)

	@property
	def is_koblitz(self):
		"""Test if a = 0 and b is sufficiently small."""
		return (self.a == 0) and (self.b < 1000)

	@property
	def order_bits(self):
		return math.log(self.n, 2)

	@property
	def field_bits(self):
		return self.p.bit_length()

	@property
	def characteristic_cmpkey(self):
		return (self.curvetype, self.a, self.b, self.n, self.h, self._domain_parameters.get("Gx"), self._domain_parameters.get("Gy"), self.p)

	def on_curve(self, point):
		lhs = (point.y * point.y) % self.p
		rhs = ((point.x ** 3) + (self.a * point.x) + self.b) % self.p
		return lhs == rhs

@EllipticCurve.register
class BinaryFieldEllipticCurve(EllipticCurve):
	"""y^2 + xy = x^3 + ax^2 + b (in F_{2^m}, reduced by irreducible poly)"""
	_DomainArgs = KwargsChecker(required_arguments = set([ "m", "poly", "a", "b", "n" ]), optional_arguments = set([ "h", "Gx", "Gy", "basis" ]))
	_CURVE_TYPE = "binary"

	def __init__(self, **domain_parameters):
		super().__init__(**domain_parameters)
		self._domain_parameters["intpoly"] = sum(1 << bit for bit in set(self.poly))

	@property
	def is_koblitz(self):
		"""Test if a, b in [ 0, 1 ]."""
		return (self.a in  [ 0, 1 ]) and (self.b in [ 0, 1 ])

	@property
	def order_bits(self):
		return math.log(self.n, 2)

	@property
	def field_bits(self):
		return self.m

	@property
	def int_poly(self):
		return sum(1 << exponent for exponent in self.poly)

	@property
	def characteristic_cmpkey(self):
		return (self.curvetype, self.a, self.b, self.n, self.h, self._domain_parameters.get("Gx"), self._domain_parameters.get("Gy"), self.m, tuple(sorted(set(self.poly))))

	def on_curve(self, point):
		lhs = NumberTheory.binpoly_reduce(NumberTheory.cl_mul(point.y, point.y) ^ NumberTheory.cl_mul(point.x, point.y), self.intpoly)
		rhs = NumberTheory.binpoly_reduce(NumberTheory.cl_mul(NumberTheory.cl_mul(point.x, point.x), point.x) ^ NumberTheory.cl_mul(NumberTheory.cl_mul(self.a, point.x), point.x) ^ self.b, self.intpoly)
		return lhs == rhs

@EllipticCurve.register
class TwistedEdwardsEllipticCurve(EllipticCurve):
	"""ax^2 + y^2 = 1 + dx^2 y^2 (mod p)"""
	_DomainArgs = KwargsChecker(required_arguments = set([ "p", "a", "d", "element_octet_cnt", "expand_bitwise_and", "expand_bitwise_or", "expand_hashfnc" ]), optional_arguments = set([ "Gx", "Gy", "expand_hashlen" ]))
	_CURVE_TYPE = "twisted_edwards"

	def __init__(self, **domain_parameters):
		super().__init__(**domain_parameters)
		self._sqrt_neg1_modp = pow(2, (self.p - 1) // 4, self.p)

	@property
	def element_octet_cnt(self):
		return self._domain_parameters["element_octet_cnt"]

	@property
	def order_bits(self):
		return NumberTheory.hamming_weight(self.expand_bitwise_and & ~self.expand_bitwise_or)

	@property
	def field_bits(self):
		return self.p.bit_length()

	@property
	def neutral_point(self):
		return self.point(0, 1)

	def on_curve(self, point):
		lhs = ((self.a * point.x * point.x) + (point.y * point.y)) % self.p
		rhs = 1 + (self.d * point.x * point.x * point.y * point.y) % self.p
		return lhs == rhs

	def point_addition(self, P, Q):
		x = (P.x * Q.y + Q.x * P.y) % self.p
		x = (x * NumberTheory.modinv(1 + self.d * P.x * Q.x * P.y * Q.y, self.p)) % self.p
		y = (P.y * Q.y - self.a * P.x * Q.x)  % self.p
		y = (y * NumberTheory.modinv(1 - self.d * P.x * Q.x * P.y * Q.y, self.p)) % self.p
		return self.point(x, y)

	def encode_point(self, point):
		encoded_point = bytearray(point.y.to_bytes(length = self.element_octet_cnt, byteorder = "little"))
		encoded_point[-1] |= (point.x & 1) << 7
		return bytes(encoded_point)

	def decode_point(self, serialized_point):
		assert(isinstance(serialized_point, bytes))
		if len(serialized_point) != self.element_octet_cnt:
			raise InvalidInputException("Do not know how to decode %s point. Expected %d octets, but got %d." % (str(self), self.element_octet_cnt, len(serialized_point)))

		serialized_point = bytearray(serialized_point)
		x_lsb = (serialized_point[-1] >> 7) & 1
		serialized_point[-1] &= 0x7f
		y = int.from_bytes(serialized_point, byteorder = "little")

		if y >= self.p:
			raise InvalidInputException("y coordinate of point must be smaller than p.")

		# x^2 = (1 - y^2) / (a - dy^2)
		x2 = (1 - y * y) % self.p
		x2 *= NumberTheory.modinv(self.a - self.d * y * y, self.p)
		(x_pos, x_neg) = NumberTheory.sqrt_mod_p(x2, self.p)

		if x_lsb == 0:
			x = x_pos
		else:
			x = x_neg

		point = self.point(x, y)
		return point

	def expand_secret(self, serialized_unhashed_secret):
		assert(isinstance(serialized_unhashed_secret, bytes))
		if len(serialized_unhashed_secret) != self.element_octet_cnt:
			raise InvalidInputException("Do not know how to decode %s secret. Expected %d octets, but got %d." % (str(self), self.element_octet_cnt, len(serialized_unhashed_secret)))
		hashfnc = hashlib.new(self.expand_hashfnc)
		hashfnc.update(serialized_unhashed_secret)
		if ("expand_hashlen" in self._domain_parameters) and (hashfnc.digest_size == 0):
			hashed_secret = hashfnc.digest(self.expand_hashlen)
		else:
			hashed_secret = hashfnc.digest()

		if ("expand_hashlen" in self._domain_parameters) and (len(hashed_secret) != self._domain_parameters["expand_hashlen"]):
			raise InvalidInputException("Expansion of secret requires %d byte hash function output, but %s provided %d bytes." % (self._domain_parameters["expand_hashlen"], str(hashfnc), len(hashed_secret)))

		scalar = int.from_bytes(hashed_secret[ : self.element_octet_cnt], byteorder = "little")
		scalar &= self.expand_bitwise_and
		scalar |= self.expand_bitwise_or
		return (scalar, self.G.scalar_mul(scalar))

	@property
	def characteristic_cmpkey(self):
		return (self.curvetype, self.a, self.d, self.p, self._domain_parameters.get("Gx"), self._domain_parameters.get("Gy"), self.element_octet_cnt, self.expand_bitwise_and, self.expand_bitwise_or, self.expand_hashfnc, self._domain_params.get("expand_hashlen"))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import pyasn1.codec.der.encoder
from pyasn1_modules import rfc3279
from x509sak.PEMDERObject import PEMDERObject

class DSAParameters(PEMDERObject):
	"""Class that allows generate deliberately broken or oddball DSA
	parameters."""
	_PEM_MARKER = "DSA PARAMETERS"
	_ASN1_MODEL = rfc3279.Dss_Parms

	@classmethod
	def create(cls, p, q, g):
		asn1 = cls._ASN1_MODEL()
		asn1["p"] = p
		asn1["q"] = q
		asn1["g"] = g
		der_data = pyasn1.codec.der.encoder.encode(asn1)
		return cls(der_data)

	@property
	def p(self):
		return int(self.asn1["p"])

	@property
	def q(self):
		return int(self.asn1["q"])

	@property
	def g(self):
		return int(self.asn1["g"])

	@property
	def L(self):
		return self.p.bit_length()

	@property
	def N(self):
		return self.q.bit_length()

	def __str__(self):
		return "DSAParameters<%d/%d>" % (self.L, self.N)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import pyasn1.type.univ
from x509sak.BijectiveDict import BijectiveDict
from x509sak.Exceptions import InvalidInputException

class OID():
	def __init__(self, oid_value):
		self._oid_value = tuple(oid_value)

	def __eq__(self, other):
		return (type(self) == type(other)) and (self._oid_value == other._oid_value)

	def __neq__(self, other):
		return not (self == other)

	def __lt__(self, other):
		return (type(self) == type(other)) and (self._oid_value < other._oid_value)

	def __hash__(self):
		return hash(self._oid_value)

	def to_asn1(self):
		return pyasn1.type.univ.ObjectIdentifier(self._oid_value)

	@classmethod
	def from_str(cls, oid_string):
		assert(isinstance(oid_string, str))
		split_string = oid_string.split(".")
		try:
			int_string = [ int(value) for value in split_string ]
		except ValueError:
			raise InvalidInputException("Cannot parse \"%s\" as a string OID." % (oid_string))
		return cls(int_string)

	@classmethod
	def from_asn1(cls, oid_asn1):
		if not isinstance(oid_asn1, pyasn1.type.univ.ObjectIdentifier):
			raise InvalidInputException("Tried to construct an OID from ASN.1 of type %s." % (oid_asn1.__class__.__name__))
		return cls.from_str(str(oid_asn1))

	def __repr__(self):
		return ".".join(str(value) for value in self._oid_value)

class OIDDB():
	"""KeySpecification algorithm OIDs."""
	KeySpecificationAlgorithms = BijectiveDict({
		OID.from_str("1.2.840.113549.1.1.1"):	"rsaEncryption",
		OID.from_str("1.2.840.10045.2.1"):		"ecPublicKey",
		OID.from_str("1.2.840.10040.4.1"):		"id-dsa",
		OID.from_str("1.3.101.110"):			"id-X25519",
		OID.from_str("1.3.101.111"):			"id-X448",
		OID.from_str("1.3.101.112"):			"id-Ed25519",
		OID.from_str("1.3.101.113"):			"id-Ed448",
	})

	"""Signature algorithm OIDs."""
	SignatureAlgorithms = BijectiveDict({
		OID.from_str("1.2.840.113549.1.1.2"):	"md2WithRsaEncryption",
		OID.from_str("1.2.840.113549.1.1.3"):	"md4WithRsaEncryption",
		OID.from_str("1.2.840.113549.1.1.4"):	"md5WithRsaEncryption",
		OID.from_str("1.2.840.113549.1.1.5"):	"sha1WithRsaEncryption",
		OID.from_str("1.2.840.113549.1.1.10"):	"RSASSA-PSS",
		OID.from_str("1.2.840.113549.1.1.11"):	"sha256WithRsaEncryption",
		OID.from_str("1.2.840.113549.1.1.12"):	"sha384WithRsaEncryption",
		OID.from_str("1.2.840.113549.1.1.13"):	"sha512WithRsaEncryption",

		OID.from_str("1.2.840.10040.4.3"):		"dsa-with-sha1",
		OID.from_str("2.16.840.1.101.3.4.3.1"):	"dsa-with-sha224",
		OID.from_str("2.16.840.1.101.3.4.3.2"):	"dsa-with-sha256",
		OID.from_str("2.16.840.1.101.3.4.3.3"):	"dsa-with-sha384",
		OID.from_str("2.16.840.1.101.3.4.3.4"):	"dsa-with-sha512",

		OID.from_str("1.2.840.10045.4.1"):		"ecdsa-with-SHA1",
		OID.from_str("1.2.840.10045.4.3.1"):	"ecdsa-with-SHA224",
		OID.from_str("1.2.840.10045.4.3.2"):	"ecdsa-with-SHA256",
		OID.from_str("1.2.840.10045.4.3.3"):	"ecdsa-with-SHA384",
		OID.from_str("1.2.840.10045.4.3.4"):	"ecdsa-with-SHA512",

		OID.from_str("1.3.14.3.2.2"):			"md4WithRsaEncryption-alt1",
		OID.from_str("1.3.14.3.2.3"):			"md5WithRsaEncryption-alt1",
		OID.from_str("1.3.14.3.2.4"):			"md5WithRsaEncryption-alt2",
		OID.from_str("1.3.14.3.2.13"):			"dsa-with-sha",
		OID.from_str("1.3.14.3.2.15"):			"shaWithRsaEncryption",
		OID.from_str("1.3.14.3.2.27"):			"sha1DSA-alt1",
		OID.from_str("1.3.14.3.2.29"):			"sha1WithRsaEncryption-alt1",

		OID.from_str("1.3.14.7.2.3.1"):			"md2WithRsaEncryption-alt1",
	})

	"""Relative Distinguished Name type component OIDs."""
	RDNTypes = BijectiveDict(value_predicate = lambda name: name.lower(), values = {
		OID.from_str("2.5.4.0"):					"objectClass",
		OID.from_str("2.5.4.1"):					"aliasedEntryName",
		OID.from_str("2.5.4.2"):					"knowledgeinformation",
		OID.from_str("2.5.4.3"):					"CN",				# common name
		OID.from_str("2.5.4.4"):					"SN",				# surname
		OID.from_str("2.5.4.5"):					"serialNumber",
		OID.from_str("2.5.4.6"):					"C",				# country
		OID.from_str("2.5.4.7"):					"L",				# locality
		OID.from_str("2.5.4.8"):					"ST",				# stateOrProvinceName
		OID.from_str("2.5.4.9"):					"street",			# streetAddress
		OID.from_str("2.5.4.10"):					"O",				# organizationName
		OID.from_str("2.5.4.11"):					"OU",				# organizationalUnitName
		OID.from_str("2.5.4.12"):					"title",
		OID.from_str("2.5.4.13"):					"description",
		OID.from_str("2.5.4.14"):					"searchGuide",
		OID.from_str("2.5.4.15"):					"businessCategory",
		OID.from_str("2.5.4.16"):					"postalAddress",
		OID.from_str("2.5.4.17"):					"postalCode",
		OID.from_str("2.5.4.18"):					"postOfficeBox",
		OID.from_str("2.5.4.19"):					"physicalDeliveryOfficeName",
		OID.from_str("2.5.4.20"):					"telephoneNumber",
		OID.from_str("2.5.4.21"):					"telexNumber",
		OID.from_str("2.5.4.22"):					"teletexTerminalIdentifier",
		OID.from_str("2.5.4.23"):					"facsimileTelephoneNumber",
		OID.from_str("2.5.4.24"):					"x121Address",
		OID.from_str("2.5.4.25"):					"internationaliSDNNumber",
		OID.from_str("2.5.4.26"):					"registeredAddress",
		OID.from_str("2.5.4.27"):					"destinationIndicator",
		OID.from_str("2.5.4.28"):					"preferredDeliveryMethod",
		OID.from_str("2.5.4.29"):					"presentationAddress",
		OID.from_str("2.5.4.30"):					"supportedApplicationContext",
		OID.from_str("2.5.4.31"):					"member",
		OID.from_str("2.5.4.32"):					"owner",
		OID.from_str("2.5.4.33"):					"roleOccupant",
		OID.from_str("2.5.4.34"):					"seeAlso",
		OID.from_str("2.5.4.35"):					"userPassword",
		OID.from_str("2.5.4.36"):					"userCertificate",
		OID.from_str("2.5.4.37"):					"cACertificate",
		OID.from_str("2.5.4.38"):					"authorityRevocationList",
		OID.from_str("2.5.4.39"):					"certificateRevocationList",
		OID.from_str("2.5.4.40"):					"crossCertificatePair",
		OID.from_str("2.5.4.41"):					"name",
		OID.from_str("2.5.4.42"):					"GN",				# givenName
		OID.from_str("2.5.4.43"):					"initials",
		OID.from_str("1.2.840.113549.1.9.1"):		"emailAddress",
		OID.from_str("0.9.2342.19200300.100.1.1"):	"UID",
		OID.from_str("0.9.2342.19200300.100.1.25"):	"DC",
	})

	"""X.509 Extension OIDs."""
	X509Extensions = BijectiveDict({
		OID.from_str("2.5.29.1"):					"oldAuthorityKeyIdentifier",
		OID.from_str("2.5.29.2"):					"oldPrimaryKeyAttributes",
		OID.from_str("2.5.29.3"):					"oldCertificatePolicies",
		OID.from_str("2.5.29.4"):					"PrimaryKeyUsageRestriction",
		OID.from_str("2.5.29.9"):					"SubjectDirectoryAttributes",
		OID.from_str("2.5.29.14"):					"SubjectKeyIdentifier",
		OID.from_str("2.5.29.15"):					"KeyUsage",
		OID.from_str("2.5.29.16"):					"PrivateKeyUsagePeriod",
		OID.from_str("2.5.29.17"):					"SubjectAlternativeName",
		OID.from_str("2.5.29.18"):					"IssuerAlternativeName",
		OID.from_str("2.5.29.19"):					"BasicConstraints",
		OID.from_str("2.5.29.20"):					"CRLNumber",
		OID.from_str("2.5.29.21"):					"ReasonCode",
		OID.from_str("2.5.29.23"):					"HoldInstructionCode",
		OID.from_str("2.5.29.24"):					"InvalidityDate",
		OID.from_str("2.5.29.27"):					"DeltaCRLIndicator",
		OID.from_str("2.5.29.28"):					"IssuingDistributionPoint",
		OID.from_str("2.5.29.29"):					"CertificateIssuer",
		OID.from_str("2.5.29.30"):					"NameConstraints",
		OID.from_str("2.5.29.31"):					"CRLDistributionPoints",
		OID.from_str("2.5.29.32"):					"CertificatePolicies",
		OID.from_str("2.5.29.33"):					"PolicyMappings",
		OID.from_str("2.5.29.35"):					"AuthorityKeyIdentifier",
		OID.from_str("2.5.29.36"):					"PolicyConstraints",
		OID.from_str("2.5.29.37"):					"ExtendedKeyUsage",
		OID.from_str("2.5.29.46"):					"FreshestCRL",
		OID.from_str("2.5.29.54"):					"X509Version3CertificateExtensionInhibitAnyPolicy",

		OID.from_str("1.3.6.1.5.5.7.1.11"):			"SubjectInformationAccess",

		OID.from_str("1.2.840.113533.7.65.0"):		"EntrustVersionExtension",

		OID.from_str("1.2.840.113549.1.9.15"):		"SMIMECapabilities",

		OID.from_str("1.3.6.1.4.1.311.20.2"):		"CertificateTemplateNameDomainController",
		OID.from_str("1.3.6.1.4.1.311.21.1"):		"CertSRVCAVersion",

		OID.from_str("1.3.6.1.4.1.11129.2.4.2"):	"CertificateTransparency",
		OID.from_str("1.3.6.1.4.1.11129.2.4.3"):	"CertificateTransparencyPrecertificatePoison",

		OID.from_str("2.23.42.7.0"):				"hashedRootKey",

		OID.from_str("1.3.6.1.5.5.7.1.1"):			"id-pe-authorityInfoAccess",
		OID.from_str("1.3.6.1.5.5.7.1.2"):			"id-pe-biometricInfo",
		OID.from_str("1.3.6.1.5.5.7.1.3"):			"id-pe-qcStatements",
		OID.from_str("1.3.6.1.5.5.7.1.12"):			"id-pe-logotype",
		OID.from_str("1.3.6.1.5.5.7.1.23"):			"NSACertificateExtension",
		OID.from_str("1.3.6.1.5.5.7.1.24"):			"TLSFeature",
		OID.from_str("1.3.6.1.5.5.7.48.1.5"):		"id-pkix-ocsp-nocheck",

		# Microsoft-specific extensions
		OID.from_str("1.3.6.1.4.1.311.21.7"):		"szOID_CERTIFICATE_TEMPLATE",
		OID.from_str("1.3.6.1.4.1.311.21.10"):		"szOID_APPLICATION_CERT_POLICIES",

		# Netscape-specific extensions
		OID.from_str("2.16.840.1.113730.1.1"):		"NetscapeCertificateType",
		OID.from_str("2.16.840.1.113730.1.2"):		"NetscapeBaseUrl",
		OID.from_str("2.16.840.1.113730.1.3"):		"NetscapeRevocationUrl",
		OID.from_str("2.16.840.1.113730.1.4"):		"NetscapeCARevocationUrl",
		OID.from_str("2.16.840.1.113730.1.5"):		"NetscapeCACRLUrl",
		OID.from_str("2.16.840.1.113730.1.6"):		"NetscapeCACertUrl",
		OID.from_str("2.16.840.1.113730.1.7"):		"NetscapeRenewalUrl",
		OID.from_str("2.16.840.1.113730.1.8"):		"NetscapeCAPolicyUrl",
		OID.from_str("2.16.840.1.113730.1.9"):		"NetscapeHomepageUrl",
		OID.from_str("2.16.840.1.113730.1.10"):		"NetscapeEntityLogo",
		OID.from_str("2.16.840.1.113730.1.11"):		"NetscapeUserPicture",
		OID.from_str("2.16.840.1.113730.1.12"):		"NetscapeSSLServerName",
		OID.from_str("2.16.840.1.113730.1.13"):		"NetscapeComment",
		OID.from_str("2.16.840.1.113730.1.14"):		"NetscapeLostPasswordUrl",
		OID.from_str("2.16.840.1.113730.1.15"):		"NetscapeCertRenewalTime",
		OID.from_str("2.16.840.1.113730.1.16"):		"NetscapeAuthorityInformationAccess",
		OID.from_str("2.16.840.1.113730.1.17"):		"NetscapeCertScopeOfUse",

		OID.from_str("2.16.840.1.113719.1.9.4.1"):	"NovellSecurityAttributes",
	})

	"""X.509 Certificate Policy OIDs."""
	X509ExtensionCertificatePolicy = BijectiveDict({
		OID.from_str("2.5.29.32.0"):				"anyPolicy",
	})

	"""X.509 Certificate Policy Qualifier OIDs."""
	X509ExtensionCertificatePolicyQualifierOIDs = BijectiveDict({
		OID.from_str("1.3.6.1.5.5.7.2.1"):			"id-qt-cps",
		OID.from_str("1.3.6.1.5.5.7.2.2"):			"id-qt-unotice",
	})

	"""Hash functions."""
	HashFunctions = BijectiveDict({
		OID.from_str("1.3.14.3.2.26"):				"sha1",
		OID.from_str("2.16.840.1.101.3.4.2.1"):		"sha256",
		OID.from_str("2.16.840.1.101.3.4.2.2"):		"sha384",
		OID.from_str("2.16.840.1.101.3.4.2.3"):		"sha512",
		OID.from_str("2.16.840.1.101.3.4.2.4"):		"sha224",
		OID.from_str("2.16.840.1.101.3.4.2.5"):		"sha512-224",
		OID.from_str("2.16.840.1.101.3.4.2.6"):		"sha512-256",
		OID.from_str("2.16.840.1.101.3.4.2.7"):		"sha3-224",
		OID.from_str("2.16.840.1.101.3.4.2.8"):		"sha3-256",
		OID.from_str("2.16.840.1.101.3.4.2.9"):		"sha3-384",
		OID.from_str("2.16.840.1.101.3.4.2.10"):	"sha3-512",
		OID.from_str("2.16.840.1.101.3.4.2.11"):	"shake128",
		OID.from_str("2.16.840.1.101.3.4.2.12"):	"shake256",
	})

	"""X.509 Extension Extended Key Usage."""
	X509ExtendedKeyUsage = BijectiveDict({
		OID.from_str("2.5.29.37.0"):				"anyExtendedKeyUsage",
		OID.from_str("1.3.6.1.5.5.7.3.1"):			"id_kp_serverAuth",
		OID.from_str("1.3.6.1.5.5.7.3.2"):			"id_kp_clientAuth",
		OID.from_str("1.3.6.1.5.5.7.3.3"):			"id_kp_codeSigning",
		OID.from_str("1.3.6.1.5.5.7.3.4"):			"id_kp_emailProtection",
		OID.from_str("1.3.6.1.5.5.7.3.5"):			"id_kp-ipsecEndSystem",
		OID.from_str("1.3.6.1.5.5.7.3.6"):			"id_kp-ipsecTunnel",
		OID.from_str("1.3.6.1.5.5.7.3.7"):			"id_kp-ipsecUser",
		OID.from_str("1.3.6.1.5.5.7.3.8"):			"id_kp_timeStamping",
		OID.from_str("1.3.6.1.5.5.7.3.9"):			"OCSPSigning",
		OID.from_str("1.3.6.1.5.5.7.3.10"):			"dvcs",
		OID.from_str("1.3.6.1.5.5.7.3.11"):			"id-kp-sbgpCertAAServerAuth",
		OID.from_str("1.3.6.1.5.5.7.3.12"):			"id-kp-scvp-responder",
		OID.from_str("1.3.6.1.5.5.7.3.13"):			"id-kp-eapOverPPP",
		OID.from_str("1.3.6.1.5.5.7.3.14"):			"id-kp-eapOverLAN",
		OID.from_str("1.3.6.1.5.5.7.3.15"):			"id-kp-scvpServer",
		OID.from_str("1.3.6.1.5.5.7.3.16"):			"id-kp-scvpClient",
		OID.from_str("1.3.6.1.5.5.7.3.17"):			"id-kp-ipsecIKE",
		OID.from_str("1.3.6.1.5.5.7.3.18"):			"id-kp-capwapAC",
		OID.from_str("1.3.6.1.5.5.7.3.19"):			"id-kp-capwapWTP",
		OID.from_str("1.3.6.1.5.5.7.3.20"):			"id-kp-sipDomain",
		OID.from_str("1.3.6.1.5.5.7.3.21"):			"secureShellClient",
		OID.from_str("1.3.6.1.5.5.7.3.22"):			"secureShellServer",
		OID.from_str("1.3.6.1.5.5.7.3.27"):			"id-kp-cmcCA",
		OID.from_str("1.3.6.1.5.5.7.3.28"):			"id-kp-cmcRA",
		OID.from_str("1.3.6.1.5.5.7.3.29"):			"id-kp-cmcArchive",
		OID.from_str("1.3.6.1.5.5.7.3.30"):			"id-kp-bgpsec-router",
		OID.from_str("1.2.203.7064.1.1.369791.1"):	"CSN369791-TLS-CLIENT",
		OID.from_str("1.2.203.7064.1.1.369791.2"):	"CSN369791-TLS-SERVER",
	})

	"""ECC field type."""
	ECFieldType = BijectiveDict({
		OID.from_str("1.2.840.10045.1.1"):			"prime-field",
		OID.from_str("1.2.840.10045.1.2"):			"characteristic-two-field",
	})

	"""ECC twofield basisType."""
	ECTwoFieldBasistype = BijectiveDict({
		OID.from_str("1.2.840.10045.1.2.3.1"):		"gnBasis",		# Gaussian basis
		OID.from_str("1.2.840.10045.1.2.3.2"):		"tpBasis",		# Trinomial basis
		OID.from_str("1.2.840.10045.1.2.3.3"):		"ppBasis",		# Pentanomial basis
	})

	"""RSAPSS Mask Generation Functions."""
	RSAPSSMaskGenerationAlgorithm = BijectiveDict({
		OID.from_str("1.2.840.113549.1.1.8"):		"mgf1",
	})

	"""Subject Information Access Method OIDs."""
	SubjectInformationAccessMethod = BijectiveDict({
		OID.from_str("1.3.6.1.5.5.7.48.1"):		"ocsp",
		OID.from_str("1.3.6.1.5.5.7.48.2"):		"caIssuers",
		OID.from_str("1.3.6.1.5.5.7.48.3"):		"timestamping",
		OID.from_str("1.3.6.1.5.5.7.48.4"):		"DVCS",							# Data Validation and Certification Server
		OID.from_str("1.3.6.1.5.5.7.48.5"):		"caRepository",
		OID.from_str("1.3.6.1.5.5.7.48.6"):		"httpCerts",
		OID.from_str("1.3.6.1.5.5.7.48.7"):		"httpCRLs",
		OID.from_str("1.3.6.1.5.5.7.48.8"):		"XKMS",							# XML Key Management Specification
		OID.from_str("1.3.6.1.5.5.7.48.9"):		"signedObjectRepository",
		OID.from_str("1.3.6.1.5.5.7.48.10"):	"rpkiManifest",					# Resource Public Key Infrastructure
		OID.from_str("1.3.6.1.5.5.7.48.11"):	"signedObject",
		OID.from_str("1.3.6.1.5.5.7.48.12"):	"CMC",							# Certificate Management over CMS
		OID.from_str("1.3.6.1.5.5.7.48.13"):	"rpkiNotify",					# Resource Public Key Infrastructure
		OID.from_str("1.3.6.1.5.5.7.48.14"):	"stirTNList",					# Secure Telephone Identity Revisited Telephony Numbers
	})



#!/usr/bin/python3
#
#	MultiCommand - Provide an openssl-style multi-command abstraction
#	Copyright (C) 2011-2021 Johannes Bauer
#
#	This file is part of pycommon.
#
#	pycommon is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pycommon is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pycommon; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#
#	File UUID 4c6b89d0-ec0c-4b19-80d1-4daba7d80967

import sys
import collections
import textwrap

from .FriendlyArgumentParser import FriendlyArgumentParser
from .PrefixMatcher import PrefixMatcher

class MultiCommand():
	RegisteredCommand = collections.namedtuple("RegisteredCommand", [ "name", "description", "parsergenerator", "action", "aliases", "visible" ])
	ParseResult = collections.namedtuple("ParseResults", [ "cmd", "args" ])

	def __init__(self, trailing_text = None):
		self._commands = { }
		self._aliases = { }
		self._cmdorder = [ ]
		self._trailing_text = trailing_text

	def register(self, commandname, description, parsergenerator, **kwargs):
		supported_kwargs = set(("aliases", "action", "visible"))
		if len(set(kwargs.keys()) - supported_kwargs) > 0:
			raise Exception("Unsupported kwarg found. Supported: %s" % (", ".join(sorted(list(supported_kwargs)))))

		if (commandname in self._commands) or (commandname in self._aliases):
			raise Exception("Command '%s' already registered." % (commandname))

		aliases = kwargs.get("aliases", [ ])
		action = kwargs.get("action")
		for alias in aliases:
			if (alias in self._commands) or (alias in self._aliases):
				raise Exception("Alias '%s' already registered." % (alias))
			self._aliases[alias] = commandname

		cmd = self.RegisteredCommand(commandname, description, parsergenerator, action, aliases, visible = kwargs.get("visible", True))
		self._commands[commandname] = cmd
		self._cmdorder.append(commandname)

	def _show_syntax(self, msg = None):
		if msg is not None:
			print("Error: %s" % (msg), file = sys.stderr)
		print("Syntax: %s [command] [options]" % (sys.argv[0]), file = sys.stderr)
		print(file = sys.stderr)
		print("Available commands:", file = sys.stderr)
		for commandname in self._cmdorder:
			command = self._commands[commandname]
			if not command.visible:
				continue
			commandname_line = command.name
			for description_line in textwrap.wrap(command.description, width = 56):
				print("    %-15s    %s" % (commandname_line, description_line))
				commandname_line = ""
		print(file = sys.stderr)
		if self._trailing_text is not None:
			for line in textwrap.wrap(self._trailing_text, width = 80):
				print(line, file = sys.stderr)
			print(file = sys.stderr)
		print("Options vary from command to command. To receive further info, type", file = sys.stderr)
		print("    %s [command] --help" % (sys.argv[0]), file = sys.stderr)

	def _raise_error(self, msg, silent = False):
		if silent:
			raise Exception(msg)
		else:
			self._show_syntax(msg)
			sys.exit(1)

	def _getcmdnames(self):
		return set(self._commands.keys()) | set(self._aliases.keys())

	def parse(self, cmdline, silent = False):
		if len(cmdline) < 1:
			self._raise_error("No command supplied.")

		# Check if we can match the command portion
		pm = PrefixMatcher(self._getcmdnames())
		try:
			supplied_cmd = pm.matchunique(cmdline[0])
		except Exception as e:
			self._raise_error("Invalid command supplied: %s" % (str(e)))

		if supplied_cmd in self._aliases:
			supplied_cmd = self._aliases[supplied_cmd]

		command = self._commands[supplied_cmd]
		parser = FriendlyArgumentParser(prog = sys.argv[0] + " " + command.name, description = command.description, add_help = False)
		command.parsergenerator(parser)
		parser.add_argument("--help", action = "help", help = "Show this help page.")
		parser.setsilenterror(silent)
		args = parser.parse_args(cmdline[1:])
		return self.ParseResult(command, args)

	def run(self, cmdline, silent = False):
		parseresult = self.parse(cmdline, silent)
		if parseresult.cmd.action is None:
			raise Exception("Should run command '%s', but no action was registered." % (parseresult.cmd.name))
		parseresult.cmd.action(parseresult.cmd.name, parseresult.args)

if __name__ == "__main__":
	mc = MultiCommand()

	def importaction(cmd, args):
		print("Import:", cmd, args)

	class ExportAction():
		def __init__(self, cmd, args):
			print("Export:", cmd, args)

	def genparser(parser):
		parser.add_argument("-i", "--infile", metavar = "filename", type = str, required = True, help = "Specifies the input text file that is to be imported. Mandatory argument.")
		parser.add_argument("--verbose", action = "store_true", help = "Increase verbosity during the importing process.")
		parser.add_argument("-n", "--world", metavar = "name", type = str, choices = [ "world", "foo", "bar" ], default = "overworld", help = "Specifies the world name. Possible options are %(choices)s. Default is %(default)s.")
	mc.register("import", "Import some file from somewhere", genparser, action = importaction, aliases = [ "ymport" ])


	def genparser(parser):
		parser.add_argument("-o", "--outfile", metavar = "filename", type = str, required = True, help = "Specifies the input text file that is to be imported. Mandatory argument.")
		parser.add_argument("--verbose", action = "store_true", help = "Increase verbosity during the importing process.")
	mc.register("export", "Export some file to somewhere", genparser, action = ExportAction)

	mc.run(sys.argv[1:])



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2019-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
from x509sak.tls.Enums import TLSVersion, CipherSuite, CompressionMethod, ECPointFormats, SupportedGroups, ExtensionType
from x509sak.tls.TLSStructs import TLSExtensionECPointFormats, TLSExtensionSupportedGroups, TLSExtensionFlag, TLSExtensionServerNameIndication
from x509sak.tls.CipherSuiteDirectory import CipherSuiteDirectory

class ClientHelloHelper():
	def __init__(self, tls_version = TLSVersion.ProtocolTLSv1_2, include_secure_renegotiation_scsv = True, cipher_suite_directory = None, allow_tls_compression = False):
		self._tls_version = tls_version
		self._include_secure_renegotiation_scsv = include_secure_renegotiation_scsv
		self._cipher_suite_directory = cipher_suite_directory
		if self._cipher_suite_directory is None:
			self._cipher_suite_directory = self._create_cipher_suite_directory()
		self._cipher_suite_directory = self._cipher_suite_directory.filter(lambda cs: not cs.is_pseudo_suite)
		self._allow_tls_compression = allow_tls_compression

	@property
	def cipher_suite_directory(self):
		return self._cipher_suite_directory

	def _create_cipher_suite_directory(self):
		csd = CipherSuiteDirectory()
		csd = csd.filter_cipher(lambda cipher: cipher.cipher in [ "AES", "ChaCha20", "DES", "Camellia", "RC4" ])
		csd = csd.filter_cipher(lambda cipher: cipher.keylen >= 128)
		csd = csd.filter_kex(lambda kex: not kex.export)
		csd = csd.filter_prf(lambda prf: prf.hashlen >= 160)
		csd = csd.filter_sig_algorithm(lambda sig_algorithm: sig_algorithm.identifier in [ "ECDSA", "RSA" ])
		return csd

	def _create_cipher_suite_list(self):
		cipher_suites = [ cs.csid for cs in self._cipher_suite_directory ]
		if self._include_secure_renegotiation_scsv:
			cipher_suites.append(CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV)
		return cipher_suites

	def _create_compression_methods(self):
		methods = [
			CompressionMethod.Null,
		]
		if self._allow_tls_compression:
			methods += [
				CompressionMethod.Deflate,
				CompressionMethod.LZS,
			]
		return methods

	def _create_supported_groups_list(self):
		return [
			SupportedGroups.X25519,
			SupportedGroups.X448,
			SupportedGroups.secp256r1,
			SupportedGroups.secp384r1,
			SupportedGroups.secp521r1,
		]

	def _create_tls_extensions(self, server_name = None):
		tls_extensions = [ ]
		if server_name is not None:
			tls_extensions.append((TLSExtensionServerNameIndication, TLSExtensionServerNameIndication.create(server_name)))
		tls_extensions += [
			(TLSExtensionECPointFormats, {
				"content": {
					"point_formats": [
						ECPointFormats.uncompressed,
						ECPointFormats.ansiX962_compressed_prime,
						ECPointFormats.ansiX962_compressed_char2,
					],
				},
			}),
			(TLSExtensionSupportedGroups, {
				"content": {
					"groups": self._create_supported_groups_list(),
				},
			}),
			(TLSExtensionFlag, { "extension_id": ExtensionType.session_ticket }),
			(TLSExtensionFlag, { "extension_id": ExtensionType.encrypt_then_mac }),
			(TLSExtensionFlag, { "extension_id": ExtensionType.extended_master_secret }),
		]
		return tls_extensions

	def create(self, server_name = None):
		client_hello = {
			"payload": {
				"handshake_protocol_version":	self._tls_version.value[1],
				"client_random":				os.urandom(32),
				"session_id":					bytes(),
				"cipher_suites":				self._create_cipher_suite_list(),
				"compression_methods":			self._create_compression_methods(),
				"extensions":					self._create_tls_extensions(server_name = server_name),
			}
		}
		return client_hello



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import math
import random
import collections
import itertools
from x509sak.Exceptions import InvalidInputException

class NumberTheory():
	"""Collection of number theoretic functions and modular arithmetic
	helpers."""
	_HammingWeightAnalysis = collections.namedtuple("HammingWeightAnalysis", [ "value", "bitlen", "hweight", "rnd_min_hweight", "rnd_max_hweight", "plausibly_random" ])
	_SMALL_PRIMES = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]

	@classmethod
	def lcm(cls, a, b):
		"""Least common multiple."""
		return (a * b) // cls.gcd(a, b)

	@classmethod
	def gcd(cls, a, b):
		"""Euclidian algorithm to compute greatest common divisor."""
		while b != 0:
			(a, b) = (b, a % b)
		return a

	@classmethod
	def egcd(cls, a, b):
		"""Extended Euclidian algorithm."""
		(s, t, u, v) = (1, 0, 0, 1)
		while True:
			if b == 0:
				return (a, s, t)

			(q, r) = divmod(a, b)
			(a, b, s, t, u, v) = (b, r, u, v, s - (q * u), t - (q * v))

	@classmethod
	def modinv(cls, a, m):
		"""Calculate modular inverse of a modulo m."""
		(g, x, y) = cls.egcd(a, m)
		if g != 1:
			raise InvalidInputException("Modular inverse of %d mod %d does not exist" % (a, m))
		else:
			return x % m

	@classmethod
	def _miller_rabin_isprime_round(cls, n):
		"""Single Miller-Rabin primality test round."""
		d = n - 1
		j = 0
		while (d % 2) == 0:
			d //= 2
			j += 1

		a = random.randint(2, n - 2)
		if pow(a, d, n) == 1:
			return True
		for r in range(j):
			if pow(a, d * (2 ** r), n) == n - 1:
				return True
		return False

	@classmethod
	def is_probable_prime(cls, p, trials = 10):
		"""Probabilistic Miller-Rabin primality test for a number p with
		'trials' Miller-Rabin rounds."""
		if p < 2:
			return False
		elif p in [ 2, 3 ]:
			return True
		for _ in range(trials):
			if not cls._miller_rabin_isprime_round(p):
				return False
		return True

	@classmethod
	def solve_crt(cls, moduli):
		"""Solve the Chinese Remainder Theorem for the given values and
		moduli."""
		# Calculate product of all moduli
		product = 1
		for modulus in moduli.keys():
			product *= modulus

		# Then determine the solution
		solution = 0
		for modulus in moduli.keys():
			if moduli[modulus] == 0:
				continue

			rem_product = product // modulus
			one_value = cls.modinv(rem_product, modulus)
			add_value = rem_product * one_value * moduli[modulus]
			solution += add_value

		return solution % product

	@classmethod
	def iter_primes(cls):
		yield 2
		p = 3
		while True:
			if cls.is_probable_prime(p):
				yield p
			p += 2

	@classmethod
	def pollard_rho(cls, n, max_iterations = None):
		"""Calculate Pollards Rho for a given number n. Returns either a factor
		of n (not necessarily prime) or None."""
		def f(x, n):
			return ((x * x) + 1) % n

		(x, y, d) = (2, 2, 1)
		iterations = 0
		while d == 1:
			if (max_iterations is not None) and (iterations == max_iterations):
				return None
			x = f(x, n)
			y = f(f(y, n), n)
			d = cls.gcd(abs(x - y), n)
			iterations += 1
		if 1 < d < n:
			return d
		else:
			assert(d == n)
			return None

	@classmethod
	def find_small_factor(cls, n):
		"""Finds a small prime factor of n by trying trial division of small
		primes."""
		for small_prime in cls._SMALL_PRIMES:
			if (n % small_prime) == 0:
				return small_prime
		return None

	@classmethod
	def factor(cls, n, max_iterations = None):
		"""Factors a compound integer."""
		factors = [ ]
		iteration = 0
		while n > 1:
			iteration += 1
			if (max_iterations is not None) and (iteration > max_iterations):
				return None

			if cls.is_probable_prime(n):
				factors.append(n)
				break

			factor = cls.find_small_factor(n)
			if factor is not None:
				factors.append(factor)
				n //= factor
				continue

			factor = cls.pollard_rho(n, max_iterations = 100)
			if factor is not None:
				factors += cls.factor(factor)
				n //= factor
				continue
		return factors

	@classmethod
	def cl_mul(cls, x, y):
		"""Carryless multiplication of x * y."""
		assert(isinstance(x, int))
		assert(isinstance(y, int))
		assert(x >= 0)
		assert(y >= 0)
		if x > y:
			(x, y) = (y, x)
		result = 0
		for bit in range(x.bit_length()):
			if x & (1 << bit):
				result ^= (y << bit)
		return result

	@classmethod
	def binpoly_reduce(cls, x, poly):
		"""Binary polynomial reduction of polynomial x against the given
		polynimal."""
		assert(isinstance(x, int))
		assert(isinstance(poly, int))
		assert(x >= 0)
		assert(poly > 0)
		while x >= poly:
			# Determine bit difference
			shift = x.bit_length() - poly.bit_length()
			x ^= (poly << shift)
		return x

	@classmethod
	def sqrt_mod_p(cls, x, p):
		"""Calculates the quadratic residue of x modulo p. Not a generic
		Tonelli-Shanks implementation, works only for p thats's 3 mod 4 or 5
		mod 8."""
		x %= p
		if (p % 4) == 3:
			sqrt = pow(x, (p + 1) // 4, p)
		elif (p % 8) == 5:
			# Two possibilities, depending on if x is a quartic residue modulo
			# p or not
			sqrt_qr = pow(x, (p + 3) // 8, p)
			if pow(sqrt_qr, 2, p) == x:
				# x is a quartic residue mod p
				sqrt = sqrt_qr
			else:
				# x is a quartic non-residue mod p
				sqrt = (sqrt_qr * pow(2, (p - 1) // 4, p)) % p
		else:
			raise NotImplementedError("Need to use Tonelli-Shanks algorithm to find quadratic residues for a p %% 8 == %d" % (p % 8))

		if ((sqrt * sqrt) % p) != (x % p):
			raise InvalidInputException("Given input value has no quadratric residue modulo p.")

		if (sqrt & 1) == 0:
			return (sqrt, p - sqrt)
		else:
			return (p - sqrt, sqrt)

	@classmethod
	def hamming_weight(cls, x):
		assert(isinstance(x, int))
		assert(x >= 0)
		weight = 0
		while x > 0:
			if x & 1:
				weight += 1
			x >>= 1
		return weight

	@classmethod
	def hamming_weight_margin(cls, bits):
		"""When a bitstring of length 'bits' is randomly generated, returns the
		Hamming weight margin in which 99.99% of all values fall. I.e., in
		99.99% of all values, the Hamming weight is between (bits / 2 - margin)
		and (bits / 2 + margin). This is a reasonable approximation that was
		calculated using fityk."""
		a0 = 8.885572
		a1 = 0.1179226
		a2 = 0.02057742
		margin = math.ceil(a0 * math.log(a1 * bits) + a2 * bits)
		if margin < 1:
			margin = 1
		return margin

	@classmethod
	def hamming_weight_analysis(cls, value, min_bit_length = None):
		"""Determines if a value is plausibly random with an error probability
		of around 0.01%."""
		bitlen = value.bit_length()
		if bitlen < 1:
			return None
		if min_bit_length is not None:
			bitlen = max(min_bit_length, bitlen)
		margin = cls.hamming_weight_margin(bitlen)
		rnd_min_hweight = (bitlen // 2) - margin
		rnd_max_hweight = (bitlen // 2) + margin
		hweight = cls.hamming_weight(value)
		plausibly_random = rnd_min_hweight <= hweight <= rnd_max_hweight
		return cls._HammingWeightAnalysis(value = value, bitlen = bitlen, hweight = hweight, rnd_min_hweight = rnd_min_hweight, rnd_max_hweight = rnd_max_hweight, plausibly_random = plausibly_random)

	@classmethod
	def asymtotic_complexity_gnfs_bits(cls, n):
		"""Return for a given value n the approximate equivalent strength in
		bits to factor using GNFS."""
		if n < 3:
			return 0
		log2_n = n.bit_length()
		log_n = log2_n * math.log(2)
		bits_security = 2.5596 * (log_n ** (1/3)) * (math.log(log_n) ** (2/3))
		bits_security = math.floor(bits_security)
		return bits_security

	@classmethod
	def randint_bits(cls, bits, two_msb_set = False):
		"""Generate a cryptographically INSECURE random integer with bit length
		'bits'. Optionally set both MSBs."""
		minval = (1 << (bits - 1))
		maxval = (1 << bits) - 1
		value = random.randint(minval, maxval)
		if two_msb_set:
			value |= 1 << (bits - 2)
		return value

	@classmethod
	def randprime_bits(cls, bits, two_msb_set = False):
		"""Generate a cryptographically INSECURE random prime with bit length
		'bits'. Optionally set both MSBs."""
		while True:
			p = cls.randint_bits(bits, two_msb_set = two_msb_set)
			p |= 1
			if cls.is_probable_prime(p):
				return p

	@classmethod
	def possible_divisors(cls, factorization):
		"""From a number of factors, gives all non-trivial divisors, including
		the trivial 1 and product(factorization)."""
		ctr = collections.Counter(factorization)
		factorization = sorted((factor, count) for (factor, count) in ctr.items())
		generators = tuple(range(count + 1) for (factor, count) in factorization)
		for exponents in itertools.product(*generators):
			product = 1
			for factor in ((factor ** exponent) for (exponent, (factor, max_exponent)) in zip(exponents, factorization)):
				product *= factor
			yield product

	@classmethod
	def isqrt(cls, value):
		if value < 0:
			raise InvalidInputException("Cannot return isqrt(%d)" % (value))
		elif value < 2:
			return value
		else:
			small = cls.isqrt(value >> 2) << 1
			large = small + 1
			if (large * large) > value:
				return small
			else:
				return large



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import socket
import time
from x509sak.tls.Enums import TLSVersion, ContentType
from x509sak.tls.TLSStructs import RecordLayerPkt
from x509sak.tls.TLSMessageDecoder import TLSMessageDecoder
from x509sak.tls.RXBuffer import RXBuffer

class TLSConnectionInterruptedException(Exception): pass

class TLSConnectionTransport():
	def __init__(self, sock):
		self._sock = sock
		self._rxbuffer = RXBuffer()

	def send(self, data):
		return self._sock.send(data)

	def recvchunk(self, length):
		chunk = self._sock.recv(length)
		self._rxbuffer += chunk
		return self._rxbuffer.get_all()

	def recv(self, length):
		self._rxbuffer.read_from(source = lambda buf: self._sock.recv(length - len(buf)), until = lambda buf: buf.have_bytes(length))
		return self._rxbuffer.get(length)

	def recvline(self):
		self._rxbuffer.read_from(source = lambda buf: self._sock.recv(1024), until = lambda buf: buf.have_line())
		return self._rxbuffer.getline()

	def close(self):
		self._sock.shutdown(socket.SHUT_RDWR)
		self._sock.close()

class TLSClientConnection():
	def __init__(self, transport, tls_version):
		assert(isinstance(tls_version, TLSVersion))
		self._transport = transport
		self._tls_version = tls_version
		self._decoder = TLSMessageDecoder(side = "server")
		self._decoder.add_hook("alert", self._msg_alert)
		self._running = True

	@property
	def transport(self):
		return self._transport

	@property
	def decoder(self):
		return self._decoder

	def send(self, content_type, message):
		assert(isinstance(content_type, ContentType))
		assert(isinstance(message, (bytes, bytearray)))
		record_layer_pkt = {
			"content_type":				content_type,
			"record_layer_version":		self._tls_version.value[0],
			"payload":					message,
		}
		frame = RecordLayerPkt.pack(record_layer_pkt)
		return self._transport.send(frame)

	def send_handshake(self, message):
		return self.send(ContentType.Handshake, message)

	def _msg_alert(self, alert):
		print("Alert: %s" % (alert))
		self._cancel()

	def _cancel(self):
		self._running = False

	def receive(self):
		self._running = True
		while self._running:
			data = self._transport.recvchunk(1024)
			if len(data) != 0:
				self._decoder.put(data)
			else:
				time.sleep(0.1)
		self._transport.close()

	@classmethod
	def tcp_connect(cls, tls_version, servername, port = 443, timeout = 10.0):
		sock = socket.create_connection((servername, port), timeout = timeout)
		transport = TLSConnectionTransport(sock)
		return cls(tls_version = tls_version, transport = transport)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
from x509sak.tests import BaseTest, ResourceFileLoader

class CmdLineTestsHashPart(BaseTest):
	def test_simple_hash(self):
		with ResourceFileLoader("certs/ok/johannes-bauer.com.pem") as infile:
			result = self._run_x509sak([ "hashpart", "-l", "3", "-h", "md5", infile ])
			self.assertIn("cfab1ba8c67c7c838db98d666f02a132", result.stdout_text)		# md5("--")
			lines = result.stdout_text.split("\n")[:-1]
			self.assertEqual(len(lines), 6)

	def test_hash_search1(self):
		with ResourceFileLoader("certs/ok/johannes-bauer.com.pem") as infile:
			result = self._run_x509sak([ "hashpart", "-o", "20", "-l", "20", "-s", "af620e0", "-h", "md5", infile ])
			self.assertIn("602223a041eaf620e00527d2c0a8cb31", result.stdout_text)
			lines = result.stdout_text.split("\n")[:-1]
			self.assertEqual(len(lines), 1)

	def test_hash_search2(self):
		with ResourceFileLoader("certs/ok/johannes-bauer.com.pem") as infile:
			result = self._run_x509sak([ "hashpart", "-o", "100", "-l", "20", "-s", "0000", "-h", "sha1", "-h", "shake_256", infile ])
			self.assertIn("b2736f6e448f6b6971a1e8ad5000009bf48752d1", result.stdout_text)
			self.assertIn("e96fe7bce088755d56506ab6238fd4c8a3a2cda6e9ddf9af58a0c79bebce02815c26731cad30bf0000b72acd5b5b5842", result.stdout_text)
			lines = result.stdout_text.split("\n")[:-1]
			self.assertEqual(len(lines), 2)

	def test_hash_der_cert(self):
		cert = self._load_crt("ok/pubkey_sig_ed25519")
		with tempfile.NamedTemporaryFile(prefix = "cert_", suffix = ".der") as certfile:
			certfile.write(cert.der_data)
			certfile.flush()
			result = self._run_x509sak([ "hashpart", "-s", "6403FEE102060D6C6907473A9115583AD83C9B90", certfile.name ])
			self.assertIn("sha1 0x94 0xb4 0x20 6403fee102060d6c6907473a9115583ad83c9b90", result.stdout_text)
			lines = result.stdout_text.split("\n")[:-1]
			self.assertEqual(len(lines), 1)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyasn1_modules import rfc3279
from x509sak.AlgorithmDB import SignatureAlgorithms, SignatureFunctions, Cryptosystems
from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate.Validator import ValidationJudgement
from x509sak.estimate import JudgementCode, Commonness, Compatibility
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements, RFCReference
from x509sak.estimate.Validator import ValidationIssue
from x509sak.NumberTheory import NumberTheory
from x509sak.RSAPSSParameters import RSAPSSParameters
from x509sak.estimate.DERValidator import DERValidator
from x509sak.Tools import ASN1Tools
from x509sak import ASN1Models

@BaseEstimator.register
class SignatureSecurityEstimator(BaseEstimator):
	_ALG_NAME = "sig"
	_DER_VALIDATOR_ECDSA_SIGNATURE = DERValidator.create_inherited("X509Cert_Signature_ECDSA", validation_subject = "ECDSA signature", specific_judgements = {
		("Enc_DER_EncodingIssues_Malformed_NonDEREncoding", "Enc_DER_EncodingIssues_Malformed_Undecodable", "Enc_DER_EncodingIssues_Malformed_UnexpectedType"): ValidationJudgement(standard = RFCReference(rfcno = 3279, sect = "2.2.3", verb = "SHALL", text = "To easily transfer these two values as one signature, they MUST be ASN.1 encoded using the following ASN.1 structure:")),
	})
	_DER_VALIDATOR_DSA_SIGNATURE = DERValidator.create_inherited("X509Cert_Signature_DSA", validation_subject = "DSA signature", specific_judgements = {
		("Enc_DER_EncodingIssues_Malformed_NonDEREncoding", "Enc_DER_EncodingIssues_Malformed_Undecodable", "Enc_DER_EncodingIssues_Malformed_UnexpectedType"): ValidationJudgement(standard = RFCReference(rfcno = 3279, sect = "2.2.2", verb = "SHALL", text = "To easily transfer these two values as one signature, they SHALL be ASN.1 encoded using the following ASN.1 structure:")),
	})

	_DER_VALIDATOR_RSAPSS_PARAMETERS = DERValidator(validation_subject = "RSA/PSS parameters", recognized_issues = {
		"Enc_DER_EncodingIssues_Malformed_NonDEREncoding":		ValidationIssue(code = JudgementCode.X509Cert_PublicKey_RSA_RSAPSS_Parameters_Malformed_NonDEREncoding),
		"Enc_DER_EncodingIssues_Malformed_UnexpectedType":		ValidationIssue(code = JudgementCode.X509Cert_PublicKey_RSA_RSAPSS_Parameters_Malformed_UnexpectedType),
	})

	def _analyze_rsa_pss_signature_params(self, signature_alg_params):
		judgements = SecurityJudgements()
		asn1_details = ASN1Tools.safe_decode(signature_alg_params, asn1_spec = ASN1Models.RSASSA_PSS_Params())
		judgements += self._DER_VALIDATOR_RSAPSS_PARAMETERS.validate(asn1_details)
		if asn1_details.asn1 is not None:
			rsapss = RSAPSSParameters.from_asn1(asn1_details.asn1)
			if rsapss.hash_algorithm is None:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_HashFunction_Unknown, "Certificate has unknown hash function for use in RSA-PSS, OID %s. Cannot make security determination for that part." % (rsapss.hash_algorithm_oid), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.LIMITED_SUPPORT)

			if rsapss.mask_algorithm is None:
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_RSA_PSS_UnknownMaskFunction, "Certificate has unknown mask function for use in RSA-PSS, OID %s." % (rsapss.mask_algorithm_oid), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.LIMITED_SUPPORT)

			if rsapss.mask_hash_algorithm is None:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_HashFunction_Unknown, "Certificate has unknown mask hash function for use in RSA-PSS, OID %s." % (rsapss.mask_hash_algorithm_oid), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.LIMITED_SUPPORT)

			if rsapss.trailer_field_value is None:
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_RSA_PSS_UnknownTrailerField, "Certificate has unknown trailer field for use in RSA-PSS, trailer field ID %d." % (rsapss.trailer_field), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.LIMITED_SUPPORT)

			if (rsapss.hash_algorithm is not None) and (rsapss.mask_hash_algorithm is not None) and (rsapss.hash_algorithm != rsapss.mask_hash_algorithm):
				standard = RFCReference(rfcno = 3447, sect = "8.1", verb = "RECOMMEND", text = "Therefore, it is recommended that the EMSA-PSS mask generation function be based on the same hash function.")
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_RSA_PSS_MultipleHashFunctions, "RSA-PSS uses hash function %s for hashing, but %s for masking. This is discouraged." % (rsapss.hash_algorithm.name, rsapss.mask_hash_algorithm.name), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.LIMITED_SUPPORT, standard = standard)

			if rsapss.salt_length < 0:
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_RSAPSS_InvalidSaltLength, "Certificate has negative salt length for use in RSA-PSS, %d bytes specified." % (rsapss.salt_length), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION, bits = 0)
			elif rsapss.salt_length == 0:
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_RSA_PSS_NoSaltUsed, "RSA-PSS does not use any salt.", commonness = Commonness.HIGHLY_UNUSUAL)
			elif rsapss.salt_length < 16:
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_RSA_PSS_ShortSaltUsed, "RSA-PSS uses a comparatively short salt value of %d bits." % (rsapss.salt_length * 8), commonness = Commonness.UNUSUAL)
			else:
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_RSA_PSS_SaltLengthInBytes, "RSA-PSS uses a salt of %d bits." % (rsapss.salt_length * 8))
			return (rsapss.hash_algorithm, judgements)
		else:
			return (None, judgements)

	def _determine_hash_function(self, signature_alg, signature_alg_params):
		hash_fnc = None
		judgements = SecurityJudgements()
		if signature_alg == SignatureAlgorithms.RSASSA_PSS:
			(hash_fnc, new_judgements) = self._analyze_rsa_pss_signature_params(signature_alg_params)
			judgements += new_judgements
		else:
			judgements += SecurityJudgement(JudgementCode.X509sakIssues_AnalysisNotImplemented, "Cannot determine hash function for signature algorithm %s. This might be a shortcoming of x509sak; please report the certificate in question to the developers." % (signature_alg.name))

		return (hash_fnc, judgements)

	def analyze(self, signature_alg_oid, signature_alg_params, signature, root_cert = None):
		judgements = SecurityJudgements()
		signature_alg = SignatureAlgorithms.lookup("oid", signature_alg_oid)
		if signature_alg is None:
			judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_Function_Unknown, "Certificate has unknown signature algorithm with OID %s. Cannot make security determination." % (signature_alg_oid), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.LIMITED_SUPPORT)
			result = {
				"name":			str(signature_alg_oid),
				"pretty":		str(signature_alg_oid),
				"security":		judgements,
			}
			return result

		if isinstance(signature_alg.value.oid, (tuple, list)):
			# Have more than one OID for this
			if signature_alg.value.oid[0] != signature_alg_oid:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_Function_DeprecatedOID, "Signature algorithm uses alternate OID %s for algorithm %s. Preferred OID would be %s." % (signature_alg_oid, signature_alg.name, signature_alg.value.oid[0]), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.LIMITED_SUPPORT)

		if signature_alg.value.hash_fnc is not None:
			# Signature algorithm already implies a concrete hash function, already done.
			hash_fnc = signature_alg.value.hash_fnc
		else:
			# Signature algorithms depends and is not implied.
			(hash_fnc, new_judgements) = self._determine_hash_function(signature_alg, signature_alg_params)
			judgements += new_judgements

		if signature_alg.value.sig_fnc == SignatureFunctions.ecdsa:
			# Decode ECDSA signature
			asn1_details = ASN1Tools.safe_decode(signature, asn1_spec = rfc3279.ECDSA_Sig_Value())
			judgements += self._DER_VALIDATOR_ECDSA_SIGNATURE.validate(asn1_details)
			if (asn1_details.asn1 is not None) and (root_cert is not None):
				if root_cert.pubkey.pk_alg.value.cryptosystem == Cryptosystems.ECC_ECDSA:
					# Check that this is really a potential parent CA certificate
					ca_curve = root_cert.pubkey.curve
					hweight_analysis = NumberTheory.hamming_weight_analysis(int(asn1_details.asn1["r"]), min_bit_length = ca_curve.field_bits)
					if not hweight_analysis.plausibly_random:
						judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_ECDSA_R_BitBiasPresent, "Hamming weight of ECDSA signature R parameter is %d at bitlength %d, but expected a weight between %d and %d when randomly chosen; this is likely not coincidential." % (hweight_analysis.hweight, hweight_analysis.bitlen, hweight_analysis.rnd_min_hweight, hweight_analysis.rnd_max_hweight), commonness = Commonness.HIGHLY_UNUSUAL)
					hweight_analysis = NumberTheory.hamming_weight_analysis(int(asn1_details.asn1["s"]), min_bit_length = ca_curve.field_bits)
					if not hweight_analysis.plausibly_random:
						judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_ECDSA_S_BitBiasPresent, "Hamming weight of ECDSA signature S parameter is %d at bitlength %d, but expected a weight between %d and %d when randomly chosen; this is likely not coincidential." % (hweight_analysis.hweight, hweight_analysis.bitlen, hweight_analysis.rnd_min_hweight, hweight_analysis.rnd_max_hweight), commonness = Commonness.HIGHLY_UNUSUAL)

		elif signature_alg.value.sig_fnc == SignatureFunctions.dsa:
			# Decode DSA signature
			asn1_details = ASN1Tools.safe_decode(signature, asn1_spec = rfc3279.Dss_Sig_Value())
			judgements += self._DER_VALIDATOR_DSA_SIGNATURE.validate(asn1_details)
			if (asn1_details.asn1 is not None) and (root_cert is not None):
				if root_cert is not None:
					if root_cert.pubkey.pk_alg.value.cryptosystem == Cryptosystems.DSA:
						field_width = root_cert.pubkey.q.bit_length()

						hweight_analysis = NumberTheory.hamming_weight_analysis(int(asn1_details.asn1["r"]), min_bit_length = field_width)
						if not hweight_analysis.plausibly_random:
							judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_DSA_R_BitBiasPresent, "Hamming weight of DSA signature R parameter is %d at bitlength %d, but expected a weight between %d and %d when randomly chosen; this is likely not coincidential." % (hweight_analysis.hweight, hweight_analysis.bitlen, hweight_analysis.rnd_min_hweight, hweight_analysis.rnd_max_hweight), commonness = Commonness.HIGHLY_UNUSUAL)

						hweight_analysis = NumberTheory.hamming_weight_analysis(int(asn1_details.asn1["s"]), min_bit_length = field_width)
						if not hweight_analysis.plausibly_random:
							judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_DSA_S_BitBiasPresent, "Hamming weight of DSA signature S parameter is %d at bitlength %d, but expected a weight between %d and %d when randomly chosen; this is likely not coincidential." % (hweight_analysis.hweight, hweight_analysis.bitlen, hweight_analysis.rnd_min_hweight, hweight_analysis.rnd_max_hweight), commonness = Commonness.HIGHLY_UNUSUAL)

		result = {
			"name":				signature_alg.name,
			"sig_fnc":			self.algorithm("sig_fnc").analyze(signature_alg.value.sig_fnc),
			"security":			judgements,
		}
		if hash_fnc is not None:
			result.update({
				"pretty":		signature_alg.value.sig_fnc.value.pretty_name + " with " + hash_fnc.value.pretty_name,
				"hash_fnc":		self.algorithm("hash_fnc").analyze(hash_fnc),
			})
		else:
			result.update({
				"pretty":		 "%s with undetermined hash function" % (signature_alg.value.sig_fnc.value.pretty_name)
			})
		return result



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyasn1_modules import rfc2459
from x509sak.PEMDERObject import PEMDERObject
from x509sak.Tools import ASN1Tools

class CertificateRevocationList(PEMDERObject):
	_PEM_MARKER = "CRL"
	_ASN1_MODEL = rfc2459.CertificateList

	@property
	def this_update(self):
		return ASN1Tools.parse_datetime(str(self.asn1["tbsCertList"]["thisUpdate"].getComponent()))

	@property
	def next_update(self):
		return ASN1Tools.parse_datetime(str(self.asn1["tbsCertList"]["nextUpdate"].getComponent()))

	@property
	def crt_count(self):
		return len(self.asn1["tbsCertList"]["revokedCertificates"])

	def __str__(self):
		return "CRL<%d entries, nextUpdate = %s UTC>" % (self.crt_count, self.next_update.strftime("%Y-%m-%d %H:%M:%S"))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import mako.lookup
import mako.exceptions
from x509sak.OID import OID, OIDDB
from x509sak.BijectiveDict import BijectiveDict

class CertGeneratorHelper():
	_Exported_Extension_OIDs = BijectiveDict({
		OIDDB.X509Extensions.inverse("SubjectKeyIdentifier"):								"SKI",
		OIDDB.X509Extensions.inverse("KeyUsage"):											"KU",
		OIDDB.X509Extensions.inverse("SubjectAlternativeName"):								"SAN",
		OIDDB.X509Extensions.inverse("IssuerAlternativeName"):								"IAN",
		OIDDB.X509Extensions.inverse("BasicConstraints"):									"BC",
		OIDDB.X509Extensions.inverse("NameConstraints"):									"NC",
		OIDDB.X509Extensions.inverse("CRLDistributionPoints"):								"CRLDP",
		OIDDB.X509Extensions.inverse("CertificatePolicies"):								"CP",
		OIDDB.X509Extensions.inverse("AuthorityKeyIdentifier"):								"AKI",
		OIDDB.X509Extensions.inverse("PolicyConstraints"):									"PC",
		OIDDB.X509Extensions.inverse("PolicyMappings"):										"PM",
		OIDDB.X509Extensions.inverse("X509Version3CertificateExtensionInhibitAnyPolicy"):	"IAP",
		OIDDB.X509Extensions.inverse("ExtendedKeyUsage"):									"EKU",
		OIDDB.X509Extensions.inverse("NetscapeCertificateType"):							"NSCT",
		OIDDB.X509Extensions.inverse("SubjectInformationAccess"):							"SIA",
		OIDDB.X509Extensions.inverse("SubjectDirectoryAttributes"):							"SDA",
		OIDDB.X509Extensions.inverse("FreshestCRL"):										"FCRL",
		OIDDB.X509Extensions.inverse("id-pe-authorityInfoAccess"):							"AIA",
	})

	@classmethod
	def known_extension_oids(cls):
		return [ str(oid) for oid in cls._Exported_Extension_OIDs.keys() ]

	@classmethod
	def extension_oid_abbreviation(cls, oid):
		if oid is None:
			return None
		return cls._Exported_Extension_OIDs.get(OID.from_str(oid))

class CertGenerator():
	def __init__(self, template):
		self._template = template
		self._parameters = { }
		try:
			self._template.render(**{
				"p":						lambda name: None,
				"h":						CertGeneratorHelper,
				"declare_parameter":		self._declare_parameter,
				"export_var":				lambda x, y: None,
				"import_vars":				lambda *x: x,
				"error":					lambda *x: None,
			})
		except:
			print(mako.exceptions.text_error_template().render())

	@classmethod
	def instantiate(cls, template_name):
		base_path = os.path.dirname(__file__) + "/templates/"
		lookup = mako.lookup.TemplateLookup([ base_path + "testcases", base_path + "blocks" ], strict_undefined = True, input_encoding = "utf-8")
		template = lookup.get_template(template_name)
		return cls(template)

	@property
	def parameters(self):
		return iter(self._parameters.items())

	def get_choices(self, name):
		return self._parameters[name]

	def _get_known_extension_oids(self):
		return [ str(oid) for oid in OIDDB.X509Extensions.keys() ]

	def _declare_parameter(self, name, choices):
		self._parameters[name] = choices

	def render(self, parameters):
		missing_keys = set(self._parameters.keys()) - set(parameters.keys())
		if len(missing_keys) > 0:
			raise Exception("Need to supply: %s" % (", ".join(sorted(missing_keys))))

		exported_vars = dict(parameters)
		def export_var(varname, value):
			exported_vars[varname] = value

		def import_vars(*names):
			if len(names) == 1:
				return exported_vars[names[0]]
			else:
				return [ exported_vars[name] for name in names ]

		def error(*args):
			raise Exception(*args)

		result = self._template.render(**{
			"p":						parameters.get,
			"h":						CertGeneratorHelper,
			"declare_parameter":		lambda x, y: None,
			"export_var":				export_var,
			"import_vars":				import_vars,
			"error":					error,
		})
		return (exported_vars["filename"], result)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import bisect

class IntervalConstraintException(Exception): pass
class IntervalIdenticalException(IntervalConstraintException): pass
class IntervalOverlapsException(IntervalConstraintException): pass

class IntervalRelation(enum.IntEnum):
	Disjunct = 0
	Identical = 1
	Overlapping = 2
	Contained = 3
	Container = 4

class Interval():
	def __init__(self, begin, end, data = None):
		"""'begin' is inclusive, 'end' is exclusive."""
		assert(isinstance(begin, int))
		assert(isinstance(end, int))
		assert(end > begin)
		self._begin = begin
		self._end = end
		self._data = data

	@classmethod
	def begin_length(cls, begin, length, data = None):
		return cls(begin = begin, end = begin + length, data = data)

	@property
	def begin(self):
		return self._begin

	@property
	def end(self):
		return self._end

	@property
	def data(self):
		return self._data

	@property
	def cmpkey(self):
		return (self.begin, self.end)

	@property
	def length(self):
		return self.end - self.begin

	def enumerate_members(self):
		return range(self.begin, self.end)

	def relation_to(self, other):
		if (self.begin, self.end) == (other.begin, other.end):
			return IntervalRelation.Identical
		elif (other.begin >= self.end) or (other.end <= self.begin):
			return IntervalRelation.Disjunct
		elif (self.begin <= other.begin < self.end) and (self.begin <= other.end <= self.end):
			# self contains other
			return IntervalRelation.Container
		elif (other.begin <= self.begin < other.end) and (other.begin <= self.end <= other.end):
			# self contained in other
			return IntervalRelation.Contained
		else:
			return IntervalRelation.Overlapping

	def __eq__(self, other):
		return self.cmpkey == other.cmpkey

	def __neq__(self, other):
		return not (self == other)

	def __lt__(self, other):
		return self.cmpkey < other.cmpkey

	def __hash__(self):
		return hash(self.cmpkey)

	def __repr__(self):
		return "Interval(%d, %d)" % (self.begin, self.end)

class Intervals():
	def __init__(self, allow_overlapping = True, allow_identical = True):
		self._allow_overlapping = allow_overlapping
		self._allow_identical = allow_identical
		self._intervals = [ ]

	@property
	def allow_overlapping(self):
		return self._allow_overlapping

	@property
	def allow_identical(self):
		return self._allow_identical

	def enumerate_members(self):
		for interval in self._intervals:
			yield from interval.enumerate_members()

	def add(self, interval):
		assert(isinstance(interval, Interval))
		if (not self.allow_identical) and self.relation_to_any_subinterval(interval, (IntervalRelation.Identical, )):
			raise IntervalIdenticalException("Cannot add interval %s, identical interval already present." % (interval))
		if (not self.allow_overlapping) and self.relation_to_any_subinterval(interval, (IntervalRelation.Identical, IntervalRelation.Overlapping, IntervalRelation.Container, IntervalRelation.Contained)):
			raise IntervalIdenticalException("Cannot add interval %s, some form of overlapping with present intervals." % (interval))
		bisect.insort(self._intervals, interval)

	def find_interacting(self, interval):
		assert(isinstance(interval, Interval))
		index = max(bisect.bisect(self._intervals, Interval(interval.begin, interval.begin + 1)) - 1, 0)
		while index < len(self._intervals):
			candidate = self[index]
			if (candidate.begin < interval.end) and (candidate.end > interval.begin):
				yield candidate
			elif candidate.begin >= interval.end:
				# No more matches possible
				break
			index += 1

	def relation_to_any_subinterval(self, interval, expected_relations):
		for interacting in self.find_interacting(interval):
			relation = interacting.relation_to(interval)
			if relation in expected_relations:
				return True
		return False

	def fully_contained_in_subinterval(self, interval):
		return self.relation_to_any_subinterval(interval, (IntervalRelation.Container, IntervalRelation.Identical))

	def __getitem__(self, index):
		return self._intervals[index]

	def dump(self):
		print("%d intervals total:" % (len(self._intervals)))
		for (index, interval) in enumerate(self._intervals):
			print("   %2d: %s" % (index, interval))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import hashlib
import collections
import time
from x509sak.BaseAction import BaseAction
from x509sak.Exceptions import InvalidInputException

class ActionHashPart(BaseAction):
	_SupportedHashFunction = collections.namedtuple("SupportedHashFunction", [ "name", "variable_output_length" ])

	def __init__(self, cmdname, args):
		BaseAction.__init__(self, cmdname, args)
		if len(args.hash_alg) == 0:
			hash_fncs = self.get_default_hash_fncs()
		elif ("all" in args.hash_alg):
			hash_fncs = self.get_supported_hash_fncs()
		else:
			hash_fncs = set(args.hash_alg)

		with open(args.filename, "rb") as f:
			f.seek(args.seek_offset)
			self._data = f.read(args.analysis_length)

		if len(self._args.variable_hash_length) == 0:
			hash_lengths = self.get_default_variable_hash_lengths_bits()
		else:
			hash_lengths = self._args.variable_hash_length
		if any((value % 8) != 0 for value in hash_lengths):
			raise InvalidInputException("Not all hash lengths are evenly divisible by 8: %s" % (", ".join("%d" % (value) for value in hash_lengths)))
		if any((value < 8) for value in hash_lengths):
			raise InvalidInputException("Hash length must be at least 8 bit, not all are: %s" % (", ".join("%d" % (value) for value in hash_lengths)))

		valid_search_chars = set("abcdefABCDEF0123456789")
		if (self._args.search is not None) and (len(set(self._args.search) - valid_search_chars) > 0):
			raise InvalidInputException("Search pattern may only contain hexadecimal characters, but '%s' was given." % (self._args.search))

		self._search = (self._args.search or "").lower()

		t0 = time.time()
		hash_fncs_by_name = { hashfnc.name: hashfnc for hashfnc in self.get_all_supported_hash_fncs() }
		for hash_fnc_name in sorted(hash_fncs):
			hash_fnc = hash_fncs_by_name[hash_fnc_name]
			if not hash_fnc.variable_output_length:
				self._run_hash_function(hash_fnc_name)
			else:
				for hash_len_bits in hash_lengths:
					self._run_hash_function(hash_fnc_name, output_length_bits = hash_len_bits)

		t1 = time.time()
		self._log.debug("Hash search took %.1f secs.", t1 - t0)

	def _run_hash_function(self, hash_fnc_name, output_length_bits = None):
		if output_length_bits is not None:
			output_length_bytes = output_length_bits // 8

		def perform_hashing(hash_data):
			hashfnc = hashlib.new(hash_fnc_name)
			hashfnc.update(hash_data)
			if output_length_bits is None:
				digest = hashfnc.hexdigest()
			else:
				digest = hashfnc.hexdigest(output_length_bytes)
			return digest

		if output_length_bits is None:
			full_hash_fnc_name = hash_fnc_name
		else:
			full_hash_fnc_name = "%s-%d" % (hash_fnc_name, output_length_bits)

		if self._args.max_offset is None:
			iterator = range(len(self._data))
		else:
			iterator = range(min(len(self._data), self._args.max_offset + 1))
		for offset in iterator:
			for length in range(1, len(self._data) - offset + 1):
				self._run_hash_iteration(full_hash_fnc_name, perform_hashing, offset, length)

	def _run_hash_iteration(self, hash_fnc_name, hash_function, offset, length):
		hash_data = self._data[offset : offset + length]
		hexdigest = hash_function(hash_data)
		if self._search in hexdigest:
			print("%s 0x%x 0x%x 0x%x %s" % (hash_fnc_name, offset, offset + length, length, hexdigest))

	@classmethod
	def get_default_hash_fncs(cls):
		return [ "md5", "sha1", "sha256", "sha384", "sha512" ]

	@classmethod
	def get_default_variable_hash_lengths_bits(cls):
		return [ 128, 256, 384 ]

	@classmethod
	def get_all_supported_hash_fncs(cls):
		def score(text):
			unwanted_chars = set("ABCDEFGHIJKLMNOPQRSTUVWXYZ-")
			score = 0
			for char in text:
				if char in unwanted_chars:
					score += 2

			# Prefer short over long
			score += len(text)
			return score

		constant_len_hashfnc_names = collections.defaultdict(list)
		variable_len_hashfnc_names = collections.defaultdict(list)
		for hashfnc_name in hashlib.algorithms_available:
			try:
				empty_hash = hashlib.new(hashfnc_name).digest()
				constant_len_hashfnc_names[empty_hash].append(hashfnc_name)
			except TypeError:
				empty_hash = hashlib.new(hashfnc_name).digest(16)
				variable_len_hashfnc_names[empty_hash].append(hashfnc_name)


		result = [ ]
		for (hashfnc_name_groups, variable_output_length) in [ (constant_len_hashfnc_names.values(), False), (variable_len_hashfnc_names.values(), True) ]:
			for hashfnc_names in hashfnc_name_groups:
				scored_names = sorted([ (score(name), name) for name in hashfnc_names ])
				chosen_name = scored_names[0][1]
				hash_fnc = cls._SupportedHashFunction(name = chosen_name, variable_output_length = variable_output_length)
				result.append(hash_fnc)
		return result

	@classmethod
	def get_supported_hash_fncs(cls):
		return sorted([ fnc.name for fnc in cls.get_all_supported_hash_fncs() ])



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
from x509sak.tests import BaseTest
from x509sak.WorkDir import WorkDir
from x509sak.RSAPrivateKey import RSAPrivateKey
from x509sak.PrimeDB import PrimeDB
from x509sak.NumberTheory import NumberTheory

class CmdLineTestsGenBrokenRSA(BaseTest):
	def test_create_rsa_key_no_overwrite(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			with open("broken_rsa.key", "wb"):
				pass
			self._run_x509sak([ "genbrokenrsa" ], on_failure = "exception-nopause", success_return_codes = [ 1 ])

	def test_create_rsa_key(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0xd06bda6bd4031ec96cb8023fd89fc9bb, 0xd578117dc5a445697a7c6e04e09c801f).write()
			self._run_x509sak([ "genbrokenrsa", "--bitlen", "256" ])
			key = RSAPrivateKey.read_pemfile("broken_rsa.key")[0]
			self.assertEqual(key.n.bit_length(), 256)
			self.assertEqual(key.p, 0xd06bda6bd4031ec96cb8023fd89fc9bb)
			self.assertEqual(key.q, 0xd578117dc5a445697a7c6e04e09c801f)
			self.assertEqual(key.e, 0x10001)
			self.assertEqual(key.d, 0x5a360028c4c14b78b770d19ce099e80b0a9b25ab6ae35098ce9e7cc27d08ca19)
			key.check_integrity()

	def test_carmichael_totient(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0xd06bda6bd4031ec96cb8023fd89fc9bb, 0xd578117dc5a445697a7c6e04e09c801f).write()
			self._run_x509sak([ "genbrokenrsa", "--bitlen", "256", "--carmichael-totient" ])
			key = RSAPrivateKey.read_pemfile("broken_rsa.key")[0]
			self.assertEqual(key.n.bit_length(), 256)
			self.assertEqual(key.p, 0xd06bda6bd4031ec96cb8023fd89fc9bb)
			self.assertEqual(key.q, 0xd578117dc5a445697a7c6e04e09c801f)
			self.assertEqual(key.e, 0x10001)
			self.assertEqual(key.d, 0x3504164f03e88396ab0cbc8200b8d91a19a60e66e09d11d9e69f27802917833)
			key.check_integrity()

	def test_create_close_q_rsa_key(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0xd7627ea571293d6bd1dc8d4664bc6ab1).write()
			self._run_x509sak([ "genbrokenrsa", "--bitlen", "256", "--close-q" ])
			key = RSAPrivateKey.read_pemfile("broken_rsa.key")[0]
			self.assertEqual(key.n.bit_length(), 256)
			self.assertEqual(key.p, 0xd7627ea571293d6bd1dc8d4664bc6ab1)
			self.assertEqual(key.q, key.p + 94)
			key.check_integrity()

	def test_create_inv_e_d_rsa_key(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0xd06bda6bd4031ec96cb8023fd89fc9bb, 0xd578117dc5a445697a7c6e04e09c801f).write()
			self._run_x509sak([ "genbrokenrsa", "--bitlen", "256", "--switch-e-d", "-e", "0x101" ])
			key = RSAPrivateKey.read_pemfile("broken_rsa.key")[0]
			self.assertEqual(key.n.bit_length(), 256)
			self.assertEqual(key.p, 0xd06bda6bd4031ec96cb8023fd89fc9bb)
			self.assertEqual(key.q, 0xd578117dc5a445697a7c6e04e09c801f)
			self.assertEqual(key.d, 0x101)
			key.check_integrity()

	def test_invalid_close_q(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self.assertIn(b"even modulus bitlength", self._run_x509sak([ "genbrokenrsa", "--bitlen", "129", "--close-q" ], success_return_codes = [ 1 ]).stderr)

	def test_invalid_stepping(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self.assertIn(b"greater or equal", self._run_x509sak([ "genbrokenrsa", "--bitlen", "128", "--q-stepping", "0" ], success_return_codes = [ 1 ]).stderr)

	def test_automatic_e(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0xd06bda6bd4031ec96cb8023fd89fc9bb, 0xd578117dc5a445697a7c6e04e09c801f).write()
			self._run_x509sak([ "genbrokenrsa", "--bitlen", "256", "-e", "-1" ])
			key = RSAPrivateKey.read_pemfile("broken_rsa.key")[0]
			self.assertEqual(key.n.bit_length(), 256)
			self.assertEqual(key.p, 0xd06bda6bd4031ec96cb8023fd89fc9bb)
			self.assertEqual(key.q, 0xd578117dc5a445697a7c6e04e09c801f)
			self.assertNotEqual(key.e, 0x10001)
			key.check_integrity()

	def test_verbosity(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0xd7627ea571293d6bd1dc8d4664bc6ab1).write()
			output = self._run_x509sak([ "genbrokenrsa", "--bitlen", "256", "-v", "--close-q" ]).stdout
			self.assertIn(b"p = 0xd7627ea571293d6bd1dc8d4664bc6ab1", output)

	def test_retry_q(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0xd7627ea571293d6bd1dc8d4664bc6ab1).write()
			result = self._run_x509sak([ "genbrokenrsa", "--bitlen", "256", "-v", "--close-q", "--public-exponent", "3", "-vv" ])
			self.assertIn(b"retrying", result.stderr)

	def test_retry_e(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0xd7627ea571293d6bd1dc8d4664bc6ab1).write()
			# Roughly with p = 0.75 there needs to be at least a retry. The
			# probability that it fails 10 times in a row without there being a
			# bug therefore is roughly one in a million.
			for i in range(10):
				result = self._run_x509sak([ "genbrokenrsa", "--bitlen", "256", "-f", "-v", "--close-q", "--public-exponent", "3", "-e", "-1", "-vv" ])
				if b"retrying" in result.stderr:
					break
			else:
				self.fail("In 10 tries, never was a retry with a different e value chosen. Highly improbable.")

	def test_gcd_n_phi_n(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0x1fd22b50d1e28365855635, 0x3af25062dcf148b85084f5).write()
			output = self._run_x509sak([ "genbrokenrsa", "--bitlen", "257", "--gcd-n-phi-n", "-v" ]).stdout
			key = RSAPrivateKey.read_pemfile("broken_rsa.key")[0]
			self.assertEqual(key.n.bit_length(), 257)
			self.assertEqual(key.p, 0x1fd22b50d1e28365855635)
			self.assertEqual(key.q, 0xea778f672d05715314fd556a2667dca7743e33da973)
			self.assertEqual((key.q - 1) % (2 * key.p), 0)
			self.assertNotEqual(NumberTheory.gcd(key.n, key.phi_n), 1)
			self.assertEqual(key.e, 0x10001)
			self.assertIn(b"gcd(n, phi(n)) = p", output)
			key.check_integrity()

	def test_incompatible_opts(self):
		with tempfile.NamedTemporaryFile(prefix = "config_", suffix = ".cnf") as f:
			result = self._run_x509sak([ "genbrokenrsa", "--close-q", "--gcd-n-phi-n", "--outfile", f.name ], success_return_codes = [ 1 ])
			self.assertIn(b"not allowed", result.stderr)

	def test_gcd_n_phi_n_try_again_q(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			PrimeDB().add(0x3f25fbdd02563798a3ee15, 0x3c1e53497fe2626fa6d389, 0x3bf7ae07a1892c3881ee69).write()
			result = self._run_x509sak([ "genbrokenrsa", "--bitlen", "259", "--gcd-n-phi-n", "-vv" ], success_return_codes = [ 1 ])
			self.assertIn(b"exhausted", result.stderr)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest

class PublicKeyTests(BaseTest):
	def test_rsa_pubkey(self):
		pubkey = self._load_crt_pubkey("ok/johannes-bauer-intermediate")
		self.assertEqual(pubkey.e, 65537)
		self.assertEqual(pubkey.n, 0x9cd30cf05ae52e47b7725d3783b3686330ead735261925e1bdbe35f170922fb7b84b4105aba99e350858ecb12ac468870ba3e375e4e6f3a76271ba7981601fd7919a9ff3d0786771c8690e9591cffee699e9603c48cc7eca4d7712249d471b5aebb9ec1e37001c9cac7ba705eace4aebbd41e53698b9cbfd6d3c9668df232a42900c867467c87fa59ab8526114133f65e98287cbdbfa0e56f68689f3853f9786afb0dc1aef6b0d95167dc42ba065b299043675806bac4af31b9049782fa2964f2a20252904c674c0d031cd8f31389516baa833b843f1b11fc3307fa27931133d2d36f8e3fcf2336ab93931c5afc48d0d1d641633aafa8429b6d40bc0d87dc393)

	def test_rsa_keyid(self):
		pubkey = self._load_crt_pubkey("ok/johannes-bauer-intermediate")
		self.assertEqual(pubkey.keyid(), bytes.fromhex("A84A6A63047DDDBAE6D139B7A64565EFF3A8ECA1"))

	def test_ecc_pubkey(self):
		pubkey = self._load_crt_pubkey("ok/johannes-bauer.com")
		self.assertEqual(pubkey.curve.name, "secp384r1")
		self.assertEqual(pubkey.x, 0x53559b24a6bf90e6533c957915ec2a5580a07fcb77571e798986869b446b012ff219f8c0b30628d95bc9de795a54df11)
		self.assertEqual(pubkey.y, 0x50a9a4ffecb67f74485cc85bd5218a6c06ea20f3fdbb41aae225c3789f5acdf65544a3b4d519b6ed6f09553afbc26e84)

	def test_ecc_keyid(self):
		pubkey = self._load_crt_pubkey("ok/johannes-bauer.com")
		self.assertEqual(pubkey.keyid(), bytes.fromhex("1A4AB011B05CFA57FB49028765169337F78D8EE6"))

	def test_eddsa_ed25519_keyid_openssl(self):
		pubkey = self._load_crt_pubkey("ok/pubkey_sig_ed25519")
		self.assertTrue(pubkey.curve.point(pubkey.x, pubkey.y).on_curve())
		self.assertEqual(pubkey.keyid(), bytes.fromhex("6403FEE102060D6C6907473A9115583AD83C9B90"))

	def test_eddsa_ed448_keyid_openssl(self):
		pubkey = self._load_crt_pubkey("ok/pubkey_sig_ed448")
		self.assertTrue(pubkey.curve.point(pubkey.x, pubkey.y).on_curve())
		self.assertEqual(pubkey.keyid(), bytes.fromhex("CDF4E080CF5DEA11FDC0BE0BA9F9CB677AD63CE1"))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2019-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.tls.Enums import TLSVersion, CipherSuite, CompressionMethod, ECPointFormats, SupportedGroups, ExtensionType, ContentType
from x509sak.tls.TLSStructs import ClientHelloPkt, TLSExtensionFlag, TLSExtensionServerNameIndication, TLSExtensionECPointFormats, TLSExtensionSupportedGroups, RecordLayerPkt
from x509sak.tls.DataBuffer import DataBuffer
from x509sak.tls.MessageHelper import ClientHelloHelper

class TLSStructureTestClientHello(BaseTest):
	def test_client_hello(self):
		"""Trying to exactly replicate a ClientHello packet we sniffed using
		Wireshark. Serialization and deserialization."""

		reference_packet = bytes.fromhex("1603010076010000720301ba616e70c10cd842eacbf26cca4bfc7e736ce718b18f90becff766fb7cff045e00000ac00ac014c009c01300ff0100003f0000001700150000126a6f68616e6e65732d62617565722e636f6d000b000403000102000a000c000a001d0017001e00190018002300000016000000170000")

		(tls_version_recordlayer, tls_version_handshake) = TLSVersion.ProtocolTLSv1_0.value
		client_hello = {
			"payload": {
				"handshake_protocol_version":	tls_version_handshake,
				"client_random":				bytes.fromhex("ba616e70c10cd842eacbf26cca4bfc7e736ce718b18f90becff766fb7cff045e"),
				"session_id":					bytes.fromhex(""),
				"cipher_suites": [
					CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA,
					CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA,
					CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA,
					CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA,
					CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV,
				],
				"compression_methods": [
					CompressionMethod.Null,
				],
				"extensions": [
					(TLSExtensionServerNameIndication, TLSExtensionServerNameIndication.create("johannes-bauer.com")),
					(TLSExtensionECPointFormats, {
						"content": {
							"point_formats": [
								ECPointFormats.uncompressed,
								ECPointFormats.ansiX962_compressed_prime,
								ECPointFormats.ansiX962_compressed_char2,
							],
						},
					}),
					(TLSExtensionSupportedGroups, {
						"content": {
							"groups": [
								SupportedGroups.X25519,
								SupportedGroups.secp256r1,
								SupportedGroups.X448,
								SupportedGroups.secp521r1,
								SupportedGroups.secp384r1,
							],
						},
					}),
					(TLSExtensionFlag, { "extension_id": ExtensionType.session_ticket }),
					(TLSExtensionFlag, { "extension_id": ExtensionType.encrypt_then_mac }),
					(TLSExtensionFlag, { "extension_id": ExtensionType.extended_master_secret }),
				]
			},
		}
		serialized_client_hello = ClientHelloPkt.pack(client_hello)

		record_layer_packet = {
			"content_type":				ContentType.Handshake,
			"record_layer_version":		tls_version_recordlayer,
			"payload":					serialized_client_hello,
		}
		serialized_record_layer_packet = RecordLayerPkt.pack(record_layer_packet)
		self.assertEqual(reference_packet, serialized_record_layer_packet)

		deserialized_record_layer = RecordLayerPkt.unpack(DataBuffer(reference_packet))
		self.assertEqual(record_layer_packet, deserialized_record_layer)

		deserialized_client_hello = ClientHelloPkt.unpack(DataBuffer(record_layer_packet["payload"]))
		self.assertEqual(deserialized_client_hello, client_hello)

	def test_client_hello_helper(self):
		chh = ClientHelloHelper(allow_tls_compression = True)

		client_hello = chh.create("foobar.com")
		serialized = ClientHelloPkt.pack(client_hello)
		deserialized = ClientHelloPkt.unpack(DataBuffer(serialized))
		self.assertEqual(client_hello, deserialized)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest

from x509sak.OtherModels import SignedCertificateTimestampList, SCTVersion
from x509sak.tls.Enums import SignatureAlgorithm, HashAlgorithm

class OtherModelsTests(BaseTest):
	def test_serialize_sctl(self):
		# We cannot use .timestamp() here because it depends on the locale and
		# we always want UTC.
		#ts = round(datetime.datetime(2020, 1, 1, 12, 34, 56).timestamp() * 1000)
		ts = 0x16f611a1180

		sstl = {
			"payload": [
				{ "sct":
					{
						"sct_version":		SCTVersion.v1,
						"log_id":			bytes(range(32)),
						"timestamp":		ts,
						"extensions":		bytes(),
						"DigitalSignature": {
							"hash_algorithm":	HashAlgorithm.sha256,
							"sig_algorithm":	SignatureAlgorithm.ED25519,
							"signature":		bytes(range(64)),
						},
					}
				},
			],
		}
		serialized_data = SignedCertificateTimestampList.pack(sstl)

		expect_payload = bytes.fromhex("""
			0071
				006f
					00
					000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f
					00 00 01 6f 61 1a 11 80
					0000

					04
					07
					0040
						000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f
		""".replace("\n", "").replace("\t", "").replace(" ", ""))
		self.assertEqual(expect_payload, serialized_data)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import readline
import code
import os
import atexit
import hashlib
import subprocess
import pyasn1.codec.der.decoder
import pyasn1.codec.der.encoder
import pyasn1.codec.ber.encoder
import x509sak
from x509sak import X509Certificate
from x509sak.BaseAction import BaseAction
from x509sak.Tools import PaddingTools, ASN1Tools
from x509sak.HexDump import HexDump

class DebugConsole(code.InteractiveConsole):
	def __init__(self, local_vars = None, histfile = os.path.expanduser(".dbgconsole_history")):
		code.InteractiveConsole.__init__(self, locals = local_vars)
		self.init_history(histfile)

	def init_history(self, histfile):
		readline.parse_and_bind("tab: complete")
		if hasattr(readline, "read_history_file"):
			try:
				readline.read_history_file(histfile)
			except IOError:
				pass
			atexit.register(self.save_history, histfile)

	def save_history(self, histfile):
		readline.write_history_file(histfile)

class ActionDebug(BaseAction):
	def __init__(self, cmdname, args):
		BaseAction.__init__(self, cmdname, args)
		crts = self._read_certificates(self._args.crtfile)

		self._hd = HexDump()
		local_vars = {
			"x509sak":	x509sak,
			"derdec":	lambda data: pyasn1.codec.der.decoder.decode(data)[0],
			"derenc":	pyasn1.codec.der.encoder.encode,
			"berenc":	pyasn1.codec.ber.encoder.encode,
			"crts":		crts,
			"c":		crts[0] if (len(crts) > 0) else None,
			"a":		crts[0].asn1 if (len(crts) > 0) else None,
			"unpkcs1":	PaddingTools.unpad_pkcs1,
			"decrsa":	self._decrypt_selfsigned_rsa_signature,
			"md5":		lambda payload: hashlib.md5(payload).digest(),
			"sha1":		lambda payload: hashlib.sha1(payload).digest(),
			"sha256":	lambda payload: hashlib.sha256(payload).digest(),
			"sha384":	lambda payload: hashlib.sha384(payload).digest(),
			"sha512":	lambda payload: hashlib.sha512(payload).digest(),
			"hd":		self._hd.dump,
			"ap":		self._asn1parse,
			"write_py":	self._write_py,
			"safedec":	ASN1Tools.safe_decode,
		}

		console = DebugConsole(local_vars = local_vars)
		for command in self._args.execute:
			print(">>> %s" % (command))
			console.runcode(command)
		if len(self._args.execute) > 0:
			print()
		if not self._args.no_interact:
			console.interact()

	def _read_certificate(self, filename):
		if not self._args.der:
			return X509Certificate.read_pemfile(filename)
		else:
			return [ X509Certificate.read_derfile(filename) ]

	def _read_certificates(self, filenames):
		certificates = [ ]
		for filename in filenames:
			certificates += self._read_certificate(filename)
		return certificates

	def _write_py(self, data, outfile):
		with open(outfile, "w") as f:
			for (name, value) in sorted(data.items()):
				if isinstance(value, int):
					if value < 100000:
						print("%s = %d" % (name, value), file = f)
					else:
						print("%s = %#x" % (name, value), file = f)
				else:
					print("# %s omitted (type %s)" % (name, value.__type__.__name__), file = f)

	def _decrypt_selfsigned_rsa_signature(self, certificate):
		plain = pow(int.from_bytes(certificate.signature, byteorder = "big"), certificate.pubkey.e, certificate.pubkey.n)
		length = (plain.bit_length() + 7) // 8
		return int.to_bytes(plain, length = length, byteorder = "big")

	def _asn1parse(self, data):
		proc = subprocess.Popen([ "openssl", "asn1parse", "-inform", "der", "-i" ], stdin = subprocess.PIPE)
		proc.communicate(data)
